import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def process_excel_file(file_path, sheet_name=0):
    # Load the Excel file from the specified sheet without headers
    data = pd.read_excel(file_path, sheet_name=sheet_name, header=None)

    # Find indices where 'Element' is found in column 0 to determine the start of data blocks
    data_blocks_indices = data.index[data[0] == 'Element'].tolist()

    # Define the expected columns for processed data
    expected_headers = ['Element', 'Signal', 'Type', 'Line', 'Apparent Concentration',
                        'Intensity Correction', 'k Ratio', 'Wt%', 'Wt% Sigma', 'Atomic %',
                        'Oxide', 'Oxide %', 'Oxide % Sigma', 'Number of Ions', 'Type of Ion',
                        'Standard Name', 'Factory Standard', 'Standardization Date', 'Beam Current (nA)']

    processed_data = []

    # Iterate through each data block
    for i, start in enumerate(data_blocks_indices):
        # The label for the block is in the row just before the 'Element' row
        label = data.iloc[start - 1, 0]

        # Identify the end of the block by finding the next 'Total' row after 'Element'
        total_index = data.index[(data[0] == 'Total') & (data.index > start)].min()
        end = total_index  # Set end to the row before 'Total'

        # Extract data block for current sample
        block = data.iloc[start + 1:end]  # Extract rows between 'Element' and 'Total'
        headers = ['Element'] + data.iloc[start, 1:].dropna().values.tolist()  # Get headers from the 'Element' row

        # Clean and standardize the column headers (for cases like the truncated or merged headers in the image)
        headers = [str(header).strip() for header in headers]  # Remove leading/trailing whitespace
        headers = [header.replace('\n', ' ').replace('\r', '') for header in headers]  # Clean any special characters

        # Align headers with expected headers by adding missing ones
        if len(headers) < len(expected_headers):
            headers += [None] * (len(expected_headers) - len(headers))

        # Limit block's columns to match available headers
        block.columns = headers[:block.shape[1]]

        # Prepare a row dictionary to append to the DataFrame
        row_dict = {'Sample Name': label}

        for _, row in block.iterrows():
            element = row['Element']

            for col in row.index[1:]:  # Skip the 'Element' column for key names
                # Only proceed if the column exists in the current row and value is not NaN
                if col and pd.notna(row.get(col, None)):  # Use .get() to handle missing columns gracefully
                    col_name = f"{col}_{element}"  # Use header from 'Element' row for column names
                    row_dict[col_name] = row[col]

        # Capture 'Total_wt%' and 'Total_Oxide %' from the 'Total' row if they exist
        total_row = data.iloc[total_index, :]  # Correctly access the 'Total' row
        row_dict['Total_wt%'] = total_row.get(headers.index('Wt%'), None) if 'Wt%' in headers else None
        row_dict['Total_Oxide%'] = total_row.get(headers.index('Oxide %'), None) if 'Oxide %' in headers else None

        processed_data.append(row_dict)

    return pd.DataFrame(processed_data)



def sort_columns(df2, norm = False):

    df=df2.copy()
    # Start with fixed priority columns
    priority_cols = ['Sample Name', 'Total_wt%', 'Total_Oxide%'] #, 'Cation_Sum']

    # Define exact matches for metrics and their intended order
    col_order = ['Oxide %_', 'Oxide % Sigma_', 'norm_Oxide %_', 'Wt%_', 'Wt% Sigma_']


    element_to_oxide = {
        'Si': 'SiO2', 'Ti': 'TiO2', 'Al': 'Al2O3', 'Ca': 'CaO',
        'Cr': 'Cr2O3', 'Fe': 'FeO', 'Ni': 'NiO', 'Na': 'Na2O',
        'Mg': 'MgO','Mn': 'MnO',
        'K': 'K2O', 'P': 'P2O5', 'S': 'SO3'
    }

    new_columns = {}
    for col in df.columns:
        if 'Oxide' in col:
            for element, oxide in element_to_oxide.items():
                if col.endswith(f'_{element}'):
                    new_col_name = col.replace(f'_{element}', f'_{oxide}')
                    new_columns[col] = new_col_name

    # Apply the renaming
    df.rename(columns=new_columns, inplace=True)


    # Find the columns that are going to be involved in the normalization
    desired_string = 'Oxide %_'
    matching_columns = [col for col in df.columns if desired_string in col]
    matching_column_indices = [df.columns.get_loc(col) for col in matching_columns]

    # Convert SO3 to SO2
    # if 'Oxide %_S' in list(df.keys()):
    #     df = df.rename(columns={'Oxide %_S': 'Oxide %_SO3'})
    #     df['Oxide %_SO'] = df['Oxide %_SO2']*(32.065+2*15.999)/(32.065+3*15.999)

    # normalization
    df_norm = df.iloc[:, matching_column_indices]
    df_norm = df_norm.map(lambda x: 0 if x < 0 else x)
    sum = df_norm.sum(axis = 1)
    df_norm = 100*df_norm.div(sum, axis = 0)

    df_norm = df_norm.add_suffix('_norm')
    df_norm = df_norm.add_prefix('norm_')
    df_norm = df_norm.fillna(0.0)

    df = pd.concat([df, df_norm], axis=1)

    # if norm is true, switch normalized and un-normalized analyses
    if norm is True:
        elements = set([col.split('_')[-1].split('_norm')[0] for col in df.columns if 'Oxide %_' in col])

        # Swap the data between corresponding columns
        for element in elements:
            if element != 'norm':
                oxide_col = f'Oxide %_{element}'
                norm_oxide_col = f'norm_Oxide %_{element}_norm'

                # Swap the data using a temporary variable
                temp = df[oxide_col].copy()
                df[oxide_col] = df[norm_oxide_col]
                df[norm_oxide_col] = temp


    # Collect and sort metric-specific columns using more accurate matching
    metric_cols = []
    for metric in col_order:
        # This loop will check if the columns end with the metric name preceded by an underscore
        # to ensure we match 'Oxide %_Mg' but not 'Something Oxide %_Mg'
        metric_cols.extend(sorted([col for col in df.columns if col.startswith(metric)]))


    # Collect all other columns not in priority or metric-specific lists
    other_cols = [col for col in df.columns if col not in priority_cols and col not in metric_cols]

    # Move 'Apparent Concentration' columns to the end
    apparent_concentration_cols = [col for col in other_cols if 'Apparent Concentration' in col]
    other_cols = [col for col in other_cols if col not in apparent_concentration_cols]

    # Combine all columns in the desired order
    final_cols = priority_cols + metric_cols + sorted(other_cols) + sorted(apparent_concentration_cols)

    # Debugging: Print out final column order
    # Move 'Apparent Concentration' columns to the end
    for bad_names in ['Apparent Concentration', 'Factory Standard', 'Line', 'Number of Ions']:

        apparent_concentration_cols = [col for col in df.columns if bad_names in col]
        final_cols = [col for col in final_cols if col not in apparent_concentration_cols] + apparent_concentration_cols

    df_out=df[final_cols]

    # replace column headers to make compatible with MinML

    df_out.rename(columns=lambda x: x.replace('norm_Oxide %_', ''), inplace=True)
    if norm is True:
        df_out.rename(columns=lambda x: x.replace('_norm','_unnorm'), inplace=True)

    df_out.rename(columns=lambda x: x.replace('Oxide %_', ''), inplace=True)
    df_out.rename(columns=lambda x: x.replace('Number of Ions', '#_ions_'), inplace=True)
    df_out.rename(columns=lambda x: x.replace('FeO', 'FeOt'), inplace=True)

    # Reorder DataFrame
    return df_out



# do MinML classification - assign cation totals and split into separate sheets in output excel.
def minClass(df):
    import mineralML as mm
    # prep dataframe for MinML analysis
    Oxides = ['SiO2', 'TiO2', 'Al2O3', 'Cr2O3', 'FeOt', 'MnO', 'MgO', 'CaO', 'Na2O', 'K2O']
    for o in Oxides:
        if o not in df.keys():
            df.loc[:,o] = np.zeros(len(df))

    df2=df[Oxides]
    df2=df2.fillna(0)

    df2['Mineral']='Olivine'
    df2['entity_id']=df.index

    df_pred_nn, probability_matrix = mm.predict_class_prob_nn(df2)

    # Extract key columns into main results dataframe
    # df['Predict_Mineral'] = df_pred_nn['Predict_Mineral']
    # df['Predict_Probabilty'] = df_pred_nn['Predict_Probability']
    columns = df.columns.tolist()
    if 'Predict_Mineral' in columns:
        columns.remove('Predict_Mineral')
    if 'Predict_Probability' in columns:
        columns.remove('Predict_Probability')
    columns.insert(1, 'Predict_Mineral')
    columns.insert(2, 'Predict_Probability')

    df.insert(1, 'Predict_Mineral', df2['Predict_Mineral'])
    df.insert(2, 'Predict_Probability', df2['Predict_Probability'])
    columns = df.columns.tolist()
    if 'Mineral' in columns:
        columns.remove('Mineral')


    df = df.reindex(columns = columns)

    # Oxygen_numbers = {'Amphibole': 22, 'Apatite': 4, 'Biotite': 10, 'Clinopyroxene': 6,
    #                   'Garnet': 12, 'Ilmenite': 3, 'K-Feldspar': 8, 'Magnetite': 4,
    #                   'Muscovite': 10, 'Olivine': 4, 'Orthopyroxene': 6, 'Plagioclase': 8,
    #                   'Quartz': 2, 'Rutile': 2, 'Spinel': 4, 'Tourmaline': 27, 'Zircon': 4}

    # for i in df['Predict_Mineral'].unique():
    #     df.loc[df['Predict_Mineral'] == i, 'Cation_Sum'] = Oxygen_numbers[i]*df.loc[df['Predict_Mineral'] == i, 'Cation_Sum'].astype(float)/oxygen_num

    return df



## Making nice things for the standards to paste into

ideal_cols_standards_silicates=['AnalysisDate', 'PersonName', 'SampleID',

'Total_Oxide%', 'MgO', 'SiO2', 'TiO2', 'Al2O3', 'CaO', 'MnO', 'P2O5', 'Na2O', 'K2O','FeOt', 'NiO', 'SO3', 'Cr2O3',

  'MgO_norm', 'SiO2_norm', 'TiO2_norm', 'Al2O3_norm', 'CaO_norm', 'MnO_norm', 'P2O5_norm', 'Na2O_norm', 'K2O_norm','FeOt_norm', 'NiO_norm', 'SO3_norm', 'Cr2O3_norm',

'Oxide % Sigma_MgO', 'Oxide % Sigma_SiO2', 'Oxide % Sigma_TiO2', 'Oxide % Sigma_Al2O3', 'Oxide % Sigma_CaO', 'Oxide % Sigma_MnO', 'Oxide % Sigma_P2O5',
'Oxide % Sigma_Na2O', 'Oxide % Sigma_K2O','Oxide % Sigma_FeOt', 'Oxide % Sigma_NiO', 'Oxide % Sigma_SO3', 'Oxide % Sigma_Cr2O3',

'Standard Name_Mg', 'Standardization Date_Mg','Standard Name_Si',  'Standardization Date_Si', 'Standard Name_Ti',  'Standardization Date_Ti', 'Standard Name_Al',  'Standardization Date_Al', 'Standard Name_Ca', 'Standardization Date_Ca',
 'Standard Name_Mn','Standardization Date_Mn',  'Standard Name_P','Standardization Date_P',
'Standard Name_Na','Standardization Date_Na', 'Standard Name_K', 'Standardization Date_K','Standard Name_Fe','Standardization Date_Fe', 'Standard Name_Ni',  'Standardization Date_Ni',
'Standard Name_S','Standardization Date_S', 'Standard Name_Cr','Standardization Date_Cr',
                     ]

def extract_silicate_standard_data(df_final, PersonName, StdName, AnalysisDate, StdString):
    # Set the required fields
    df_final['PersonName'] = PersonName
    df_final['StdName'] = StdName
    df_final['AnalysisDate'] = AnalysisDate

    # Create a new DataFrame with all ideal columns, initializing missing columns with empty values
    df_standard = pd.DataFrame(columns=ideal_cols_standards_silicates)

    # Fill in values from df_final where the columns match
    for col in df_final.columns:
        if col in df_standard.columns:
            df_standard[col] = df_final[col]



    # Display the new DataFrame
    #df_filtered2=df_standard.loc[df_standard['SampleID'].str.contains(StdString)]
    df_filtered2 = df_standard.loc[df_standard['SampleID'].fillna('').str.contains(StdString, na=False)]


    # Fill 'Standardization Date' columns with 'Factory' if empty
    standard_date_cols = [col for col in df_standard.columns if col.startswith('Standardization Date')]

    # Use .copy() to ensure we're working with a copy to avoid chained assignment issues
    df_filtered2 = df_filtered2.copy()

    # Fill only the 'Standardization Date' columns with 'Factory' where values are empty (NaN)
    df_filtered2[standard_date_cols] = df_filtered2[standard_date_cols].fillna('Factory')
    display(df_filtered2.head())

    return df_filtered2


# def extract_silicate_standard_data(df_final, PersonName, StdName, AnalysisDate, StdString):
#
#     df_final['PersonName']=PersonName
#     df_final['StdName']=StdName
#     df_final['AnalysisDate']=AnalysisDate
#
#     # Ensure the columns you want are in the DataFrame
#     filtered_columns = [col for col in ideal_cols_standards_silicates if col in df_final.columns]
#
#     # Create the new DataFrame
#     df_filtered = df_final[filtered_columns]
#
#     # Display the new DataFrame
#     df_filtered2=df_filtered.loc[df_filtered['SampleID'].str.contains(StdString)]
#
#     print(df_filtered2.head())
#
#
#
#     return df_filtered2





import numpy as np
import matplotlib.pyplot as plt
from functools import partial
import inspect
import warnings as w
import numbers
import pandas as pd
from Thermobar.core import *


# Liquid-only thermometry functions

## All liquid-only thermometers




def T_Put2008_eq13(P=None, *, MgO_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids: Equation 13 of Putirka et al. (2008)
    :cite:`putirka2008thermometers`

    SEE=±72 °C
    '''
    return (26.3 * MgO_Liq + 994.4 + 273.15)


def T_Shi_Test(P, *, MgO_Liq, SiO2_Liq, H2O_Liq):
    '''
    Testing the thermometer of Shi et al. (in prep)
    '''
    return (
    845.54899873 + (25.85511871*MgO_Liq)
+ (3.08839706*SiO2_Liq)
- (61.64200793*(np.log(1+H2O_Liq)))
+ (61.13449845*P/10)
)+273.15


def T_Put2008_eq14(P=None, *, Mg_Number_Liq_NoFe3, MgO_Liq,
                   FeOt_Liq, Na2O_Liq, K2O_Liq, H2O_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids: Equation 14 of Putirka et al. (2008)
    :cite:`putirka2008thermometers`

    SEE=±58 °C
    '''
    return (754 + 190.6 * Mg_Number_Liq_NoFe3 + 25.52 * MgO_Liq + 9.585 *
            FeOt_Liq + 14.87 * (Na2O_Liq + K2O_Liq) - 9.176 * H2O_Liq + 273.15)


def T_Put2008_eq15(P, *, Mg_Number_Liq_NoFe3, MgO_Liq,
                   FeOt_Liq, Na2O_Liq, K2O_Liq, H2O_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids: Equation 15 of Putirka et al. (2008).
    Pressure-dependent form of Equation 14.
    :cite:`putirka2008thermometers`

    SEE=±46 °C
    '''
    return (815.3 + 265.5 * Mg_Number_Liq_NoFe3 + 15.37 * MgO_Liq + 8.61 * FeOt_Liq +
            6.646 * (Na2O_Liq + K2O_Liq) + 39.16 * 0.1 * P - 12.83 * H2O_Liq + 273.15)


def T_Put2008_eq16(P, *, SiO2_Liq_mol_frac,
                   Al2O3_Liq_mol_frac, MgO_Liq_mol_frac):
    '''
    Liquid-only thermometer for ol-cpx-plag-saturated liquids:
    Equation 16 of Putirka et al. (2008). Adapted from Yang et al (1996).
    :cite:`putirka2008thermometers`

    SEE=±19 °C
    '''
    return (- 583 + 3141 * SiO2_Liq_mol_frac + 15779 * Al2O3_Liq_mol_frac + 1338.6 *
            MgO_Liq_mol_frac - 31440 * (Al2O3_Liq_mol_frac) * (SiO2_Liq_mol_frac) + 77.67 * 0.1 * P + 273.15)


def T_Helz1987_MgO(P=None, *, MgO_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids:
    Equation 1 (MgO thermometer) of Helz and Thornber (1987)
    :cite:``

    SEE=±10 °C
    '''
    return (20.1 * MgO_Liq + 1014 + 273.15)


def T_Shea2022_MgO(P=None, *, MgO_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids:
    Shea et al. (2022) adaptation of Helz and Thornber
    :cite:``

    SEE=±10 °C
    '''
    return (21.2 * MgO_Liq + 1017 + 273.15)

def T_Montierth1995_MgO(P=None, *, MgO_Liq):
    '''
    Liquid-only thermometer for olivine-saturated liquids:
    update of Helz and Thornber (1987) by Montrieth et al. (1995).
    :cite:``

    SEE=±10 °C
    '''
    return (23.0 * MgO_Liq + 1012 + 273.15)


def T_Helz1987_CaO(P=None, *, CaO_Liq):
    '''
    Liquid-only thermometer for glasses with Ol+Augite+Plag
    Equation 2 (CaO thermometer) of Helz and Thornber (1987)
    :cite:``

    SEE=±10 °C
    '''
    return (16.6 * CaO_Liq + 968 + 273.15)


def T_Beatt93_BeattDMg(P, *, Den_Beat93):
    '''
    Liquid-only thermometer.
    Re-arrangement of Beattie (1993) by Putirka (2008) such that
    an olivine composition isn't required (DMg ol-melt is calculated rather than measured).
    :cite:``

    '''
    return (((113.1 * 1000) / 8.3144 + (0.1 * P * 10**9 - 10**5)
            * 4.11 * (10**(-6)) / 8.3144) / Den_Beat93)


def T_Beatt93_BeattDMg_HerzCorr(P, *, Den_Beat93):
    '''
    Liquid-only thermometer. Herzberg and O'Hara (2002) correction
    to the olivine-free re-arrangment (T_Beatt93_BeattDMg) of Beattie (1993) by Putirka.
    Eliminates systematic error at high pressures
    :cite:``

    '''
    return (((113.1 * 1000 / 8.3144 + (0.0001 * (10**9) - 10**5) * 4.11 *
            (10**(-6)) / 8.3144) / Den_Beat93) + 54 * (0.1 * P) + 2 * (0.1 * P)**2)




def T_Sug2000_eq1(P=None, *, MgO_Liq_mol_frac):

    '''
    Liquid-only thermometer. Equation 1 of Sugawara et al. (2000) for olivine-saturated liquids at 0.1 MPa
    :cite:``

    '''
    return (1316 + 12.95 * MgO_Liq_mol_frac * 100)
# Coefficients form their Table 2 for olivine-liquid


def T_Sug2000_eq3_ol(P, *, MgO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 3 of Sugawara et al. (2000) for olivine-saturated liquids.
    Unlike eq 1, this is pressure-dependent
    :cite:``
    '''
    A = 1293
    B = 14.60
    C = 0.0055
    return (A + B * MgO_Liq_mol_frac * 100 + C * (P * 1000))

# Coefficients form their Table 2 for opx-liquid


def T_Sug2000_eq3_opx(P, *, MgO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 3 of Sugawara et al. (2000) for opx-saturated liquids.
    :cite:``
    '''
    A = 1324
    B = 12.22
    C = 0.0052
    return (A + B * MgO_Liq_mol_frac * 100 + C * (P * 1000))

# Coefficients form their Table 2 for cpx-liquid


def T_Sug2000_eq3_cpx(P, *, MgO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 3 of Sugawara et al. (2000) for cpx-saturated liquids.
    :cite:``
    '''
    A = 1289
    B = 12.35
    C = 0.0086
    return (A + B * MgO_Liq_mol_frac * 100 + C * (P * 1000))

# Coefficients form their Table 2 for pig-liquid


def T_Sug2000_eq3_pig(P, *, MgO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 3 of Sugawara et al. (2000) for pig-saturated liquids.
    :cite:``
    '''
    A = 1294
    B = 13.47
    C = 0.0052
    return (A + B * MgO_Liq_mol_frac * 100 + C * (P * 1000))

def T_Sug2000_eq6a_H7a(P, *, SiO2_Liq_mol_frac,
                   FeOt_Liq_mol_frac, MgO_Liq_mol_frac, CaO_Liq_mol_frac, H2O_mol_frac):
    '''
    Liquid-only thermometer. Equation 6a of Sugawara et al. (2000) for
    olivine-saturated liquids. Adds terms for CaO, SiO2 and FeO relative to equation 1 and 3.
    Included corrections for H2O given in their equation 7a.
    :cite:``
    '''
    return ((1446 - 1.440 * SiO2_Liq_mol_frac * 100 - 0.5 * FeOt_Liq_mol_frac * 100 +
            12.32 * MgO_Liq_mol_frac * 100 - 3.899 * CaO_Liq_mol_frac * 100 + 0.0043 * (P * 1000))
            -5.403*100*H2O_mol_frac)

def T_Sug2000_eq6a(P, *, SiO2_Liq_mol_frac,
                   FeOt_Liq_mol_frac, MgO_Liq_mol_frac, CaO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 6a of Sugawara et al. (2000) for
    olivine-saturated liquids. Adds terms for CaO, SiO2 and FeO relative to equation 1 and 3.
    :cite:``

    '''
    return (1446 - 1.440 * SiO2_Liq_mol_frac * 100 - 0.5 * FeOt_Liq_mol_frac * 100 +
            12.32 * MgO_Liq_mol_frac * 100 - 3.899 * CaO_Liq_mol_frac * 100 + 0.0043 * (P * 1000))


def T_Sug2000_eq6b(P, *, SiO2_Liq_mol_frac,
                   FeOt_Liq_mol_frac, MgO_Liq_mol_frac, CaO_Liq_mol_frac):
    '''
    Liquid-only thermometer. Equation 6b of Sugawara et al. (2000) for cpx-saturated liquids.
    Adds terms for CaO, SiO2 and FeO relative to equation 3
    :cite:``
    '''
    return (1202 + 1.511 * SiO2_Liq_mol_frac * 100 - 1.426 * FeOt_Liq_mol_frac * 100 +
            8.780 * MgO_Liq_mol_frac * 100 + 5.537 * CaO_Liq_mol_frac * 100 + 0.0081 * (P * 1000))

def T_Sug2000_eq6b_H7b(P, *, SiO2_Liq_mol_frac,
                   FeOt_Liq_mol_frac, MgO_Liq_mol_frac, CaO_Liq_mol_frac, H2O_mol_frac):
    '''
    Liquid-only thermometer. Equation 6b of Sugawara et al. (2000) for cpx-saturated liquids.
    Adds terms for CaO, SiO2 and FeO relative to equation 3.
    Included corrections for H2O given in their equation 7b.
    :cite:``
    '''
    return ((1202 + 1.511 * SiO2_Liq_mol_frac * 100 - 1.426 * FeOt_Liq_mol_frac * 100 +
            8.780 * MgO_Liq_mol_frac * 100 + 5.537 * CaO_Liq_mol_frac * 100 + 0.0081 * (P * 1000))
            -5.674*100*H2O_mol_frac)

def T_Put2008_eq19_BeattDMg(P, *, calcDMg_Beat93, Beat_CNML, Beat_CSiO2L, Beat_NF):
    '''
    Liquid-only thermometer. Combining terms from Beattie et al. (1993) by Putirka (2008).
    This function uses calculated DMg from Beattie, so you don't need a measured olivine composition.
    :cite:`putirka2008thermometers`
    '''
    return ((13603) + (4.943 * 10**(-7)) * ((0.1 * P)*10**9 - 10**(-5))) / (6.26 + 2 *
            np.log(calcDMg_Beat93) + 2 * np.log(1.5 * Beat_CNML) + 2 * np.log(3 * Beat_CSiO2L) - Beat_NF)



def T_Put2008_eq21_BeattDMg(P, *, calcDMg_Beat93, Na2O_Liq, K2O_Liq, H2O_Liq):
    '''
    Liquid-only thermometer (adapted from ol-liq thermometer using calc DMg from Beattie) Putirka (2008),
    equation 21 (originally Putirka et al., 2007,  Eq 2). Recalibration of Beattie (1993) to account for the
    pressure sensitivity noted by Herzberg and O'Hara (2002), and eliminates the systematic error of Beattie (1993) for hydrous compositions.
    :cite:`putirka2008thermometers`

    '''
    return (1 / ((np.log(calcDMg_Beat93) + 2.158 - 5.115 * 10**(-2) * (Na2O_Liq +
            K2O_Liq) + 6.213 * 10**(-2) * H2O_Liq) / (55.09 * 0.1 * P + 4430)) + 273.15)


def T_Put2008_eq22_BeattDMg(P, *, calcDMg_Beat93, Beat_CNML, Beat_CSiO2L, Beat_NF, H2O_Liq):
    '''
    Liquid-only thermometer (adapted from ol-liq thermometer using calc DMg from Beattie): Putirka (2008),
    equation 22 (originally Putirka et al., 2007,  Eq 4). Recalibration of Beattie (1993) to account for the pressure
    sensitivity noted by Herzberg and O'Hara (2002), and eliminates the systematic error of Beattie (1993) for hydrous compositions.
    :cite:`putirka2008thermometers`
    '''
    return ((15294.6 + 1318.8 * 0.1 * P + 2.48348 * ((0.1 * P)**2)) / (8.048 + 2.8352 * np.log(calcDMg_Beat93) + 2.097 *
            np.log(1.5 * Beat_CNML) + 2.575 * np.log(3 * Beat_CSiO2L) - 1.41 * Beat_NF + 0.222 * H2O_Liq + 0.5 * (0.1 * P)) + 273.15)

## Functions for saturation surfaces of minerals based just on liquids


def T_Molina2015_amp_sat(P=None, *, Mg_Liq_cat_frac,
                         Ca_Liq_cat_frac, Al_Liq_cat_frac):
    '''
    Amphibole-saturation thermometer from Molina et al. (2015)
    :cite:``
    '''

    return (273.15 + 107 * np.log(Mg_Liq_cat_frac) - 108 *
            np.log(Ca_Liq_cat_frac / (Ca_Liq_cat_frac + Al_Liq_cat_frac)) + 1184)

def T_Put2016_eq3_amp_sat(P=None, *, FeOt_Liq_mol_frac_hyd, CaO_Liq_mol_frac_hyd, SiO2_Liq_mol_frac_hyd,
                          TiO2_Liq_mol_frac_hyd, MgO_Liq_mol_frac_hyd, MnO_Liq_mol_frac_hyd, Al2O3_Liq_mol_frac_hyd):
    '''
    Equation 3 of Putirka et al. (2016)
    Amphibole-Liquid thermometer- temperature at which a liquid is saturated in amphibole.
    :cite:``
    '''
    return (273.15 + (24429.2) / (2.31 + 42.1 * FeOt_Liq_mol_frac_hyd
    + 32.2 * CaO_Liq_mol_frac_hyd + 2.21 * np.log(SiO2_Liq_mol_frac_hyd) -
    1.4 * np.log(TiO2_Liq_mol_frac_hyd)
    - 2.666 * np.log((FeOt_Liq_mol_frac_hyd + MgO_Liq_mol_frac_hyd) * Al2O3_Liq_mol_frac_hyd)))


def T_Put1999_cpx_sat(P, *, Mg_Liq_cat_frac, Fet_Liq_cat_frac,
                      Ca_Liq_cat_frac, Si_Liq_cat_frac, Al_Liq_cat_frac):
    '''
    Liquid-only thermometer, Putirka (1999). temperature at which a liquid is saturated
    in clinopyroxene (for a given P).
    '''
    return (10 ** 4 / (3.12 - 0.0259 * P - 0.37 * np.log(Mg_Liq_cat_frac / (Mg_Liq_cat_frac + Fet_Liq_cat_frac))
                       + 0.47 * np.log(Ca_Liq_cat_frac * (Mg_Liq_cat_frac +
                                                           Fet_Liq_cat_frac) * (Si_Liq_cat_frac)**2)
                       - 0.78 * np.log((Mg_Liq_cat_frac + Fet_Liq_cat_frac)
                                       ** 2 * (Si_Liq_cat_frac)**2)
                       - 0.34 * np.log(Ca_Liq_cat_frac * (Al_Liq_cat_frac)**2 * Si_Liq_cat_frac)))


def T_Put2008_eq34_cpx_sat(P, *, H2O_Liq, Ca_Liq_cat_frac, Si_Liq_cat_frac, Mg_Liq_cat_frac):
    '''
    Liquid-only thermometer- temperature at which a liquid is saturated in clinopyroxene (for a given P). Equation 34 of Putirka et al. (2008)
    '''
    return (10 ** 4 / (6.39 + 0.076 * H2O_Liq - 5.55 * (Ca_Liq_cat_frac * Si_Liq_cat_frac)
            - 0.386 * np.log(Mg_Liq_cat_frac) - 0.046 * P + 2.2 * (10 ** (-4)) * P**2))


def T_Beatt1993_opx(P, *, Ca_Liq_cat_frac, Fet_Liq_cat_frac, Mg_Liq_cat_frac,
                    Mn_Liq_cat_frac, Al_Liq_cat_frac, Ti_Liq_cat_frac):
    '''
    Opx-Liquid thermometer of Beattie (1993). Only uses liquid composition.
    Putirka (2008) warn that overpredicts for hydrous compositions at <1200°C, and anhydrous compositions at <1100°C
    '''
    Num_B1993 = 125.9 * 1000 / 8.3144 + \
        ((0.1 * P) * 10**9 - 10**5) * 6.5 * (10**(-6)) / 8.3144
    D_Mg_opx_li1 = (0.5 - (-0.089 * Ca_Liq_cat_frac - 0.025 * Mn_Liq_cat_frac + 0.129 * Fet_Liq_cat_frac)) / \
        (Mg_Liq_cat_frac + 0.072 * Ca_Liq_cat_frac +
         0.352 * Mn_Liq_cat_frac + 0.264 * Fet_Liq_cat_frac)
    Cl_NM = Mg_Liq_cat_frac + Fet_Liq_cat_frac + \
        Ca_Liq_cat_frac + Mn_Liq_cat_frac
    NF = (7 / 2) * np.log(1 - Al_Liq_cat_frac) + \
        7 * np.log(1 - Ti_Liq_cat_frac)
    Den_B1993 = 67.92 / 8.3144 + 2 * \
        np.log(D_Mg_opx_li1) + 2 * np.log(2 * Cl_NM) - NF
    return Num_B1993 / Den_B1993




def T_Put2005_eqD_plag_sat(P, *, Ca_Liq_cat_frac,
                           H2O_Liq, Si_Liq_cat_frac, Al_Liq_cat_frac):
    '''
    Plagioclase-Liquid saturation temperature thermometer of Putirka (2005) eq. D
    '''
    return ((10**4 / (8.759 - 6.396 * Ca_Liq_cat_frac + 0.2147 * H2O_Liq + 1.221 *
            Si_Liq_cat_frac**3 - 1.751 * 10**-2 * P - 8.043 * Al_Liq_cat_frac)))


def T_Put2008_eq26_plag_sat(P, *, Si_Liq_cat_frac,
                            Ca_Liq_cat_frac, Al_Liq_cat_frac, K_Liq_cat_frac, H2O_Liq):
    '''
    Plagioclase-Liquid saturation temperature thermometer of Putirka (2008) eq. 26. Update of P2005_eqD
    SEE=+-37C
    '''
    return (10**4 / (10.86 - 9.7654 * Si_Liq_cat_frac + 4.241 * Ca_Liq_cat_frac
                     - 55.56 * Ca_Liq_cat_frac * Al_Liq_cat_frac +
                     37.5 * K_Liq_cat_frac * Al_Liq_cat_frac
                     + 11.206 * Si_Liq_cat_frac**3 - 3.151 * 10**(-2) * (P / 10) * 10 + 0.1709 * H2O_Liq))

def T_Put2008_eq24c_kspar_sat(P, *, Al_Liq_cat_frac, Na_Liq_cat_frac,
                  K_Liq_cat_frac, Si_Liq_cat_frac, Ca_Liq_cat_frac, H2O_Liq):
    '''
    Alkali Felspar saturation thermometer of Putirka (2008) eq. 24b.
    SEE=+-23 C (Calibration data)
    SEE=+-25 C (All data)

    '''
    return (10**4 / (14.6 + 0.055 * H2O_Liq - 0.06 * (P / 10) / 10 - 99.6 * Na_Liq_cat_frac * Al_Liq_cat_frac
    - 2313 * Ca_Liq_cat_frac *Al_Liq_cat_frac + 395 * K_Liq_cat_frac * Al_Liq_cat_frac
    - 151 * K_Liq_cat_frac * Si_Liq_cat_frac + 15037 * Ca_Liq_cat_frac**2))





## Listing them all to check for invalid inputs- add new ones into this list so they become recognised.

Liquid_only_funcs = {T_Put2008_eq13, T_Put2008_eq14, T_Put2008_eq15, T_Put2008_eq16,
T_Helz1987_MgO, T_Montierth1995_MgO, T_Helz1987_CaO, T_Beatt93_BeattDMg,
T_Beatt93_BeattDMg_HerzCorr, T_Sug2000_eq1, T_Sug2000_eq3_ol, T_Sug2000_eq3_opx,
T_Sug2000_eq3_cpx, T_Sug2000_eq3_pig,
T_Sug2000_eq6a, T_Sug2000_eq6a_H7a, T_Sug2000_eq6b, T_Sug2000_eq6b_H7b, T_Put2008_eq19_BeattDMg, T_Put2008_eq21_BeattDMg,
T_Put2008_eq22_BeattDMg, T_Molina2015_amp_sat, T_Put2016_eq3_amp_sat,
T_Put2008_eq34_cpx_sat,
T_Put1999_cpx_sat, T_Put2008_eq26_plag_sat, T_Put2005_eqD_plag_sat, T_Put2008_eq24c_kspar_sat, T_Beatt1993_opx, T_Shi_Test, T_Shea2022_MgO} # put on outside

Liquid_only_funcs_by_name = {p.__name__: p for p in Liquid_only_funcs}


def calculate_liq_only_temp(*, liq_comps, equationT, P=None, H2O_Liq=None, print=False):

    '''
    Liquid-only thermometery. Returns a temperature in Kelvin.

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    equationT: str
        If has _sat at the end, represents the saturation surface of that mineral.

        Equations from Putirka et al. (2016).
            | T_Put2016_eq3_amp_sat (saturation surface of amphibole)

        Equations from Putirka (2008) and older studies:

            | T_Put2008_eq13
            | T_Put2008_eq14
            | T_Put2008_eq15
            | T_Put2008_eq16
            | T_Put2008_eq34_cpx_sat
            | T_Put1999_cpx_sat
            * Following 3 thermometers are adaptations of olivine-liquid thermometers with  DMg calculated using Beattie 1993,
            This means you can use them without knowing an olivine composition. ocan be applied when you haven't measured an olivine composiiton.
            | T_Put2008_eq19_BeattDMg
            | T_Put2008_eq21_BeattDMg
            | T_Put2008_eq22_BeattDMg

        Equations from Sugawara (2000):

            | T_Sug2000_eq1
            | T_Sug2000_eq3_ol
            | T_Sug2000_eq3_opx
            | T_Sug2000_eq3_cpx
            | T_Sug2000_eq3_pig
            | T_Sug2000_eq6a
            | T_Sug2000_eq6b

        Equations from Helz and Thornber (1987):

            | T_Helz1987_MgO
            | T_Helz1987_CaO

        Equation from Molina et al. (2015)

            | T_Molina2015_amp_sat

        Equation from Montrieth 1995
           | T_Montierth1995_MgO

        Equation from Beattie (1993)
           | T_Beatt1993_opx

        Equation from Shea et al. (2022)
           | T_Shea2022_MgO

    P: float, int, pandas.Series, str  ("Solve")
        Pressure in kbar
        Only needed for P-sensitive thermometers.
        If enter P="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    H2O_Liq: optional.
        If None, uses H2O_Liq column from input.
        If int, float, pandas.Series, uses this instead of H2O_Liq Column


    Returns
    -------
    pandas series
       Temperature in K

    '''


# This checks if your function is one of the accepted liquid equations
    try:
        func = Liquid_only_funcs_by_name[equationT]
    except KeyError:
        raise ValueError(f'{equationT} is not a valid equation') from None
    sig=inspect.signature(func)


    if isinstance(P, pd.Series):
        if len(P) != len(liq_comps):
            raise ValueError('The panda series entered for pressure isnt the same length as the dataframe of liquid compositions')


    liq_comps_c = liq_comps.copy()

    if H2O_Liq is not None:
        liq_comps_c['H2O_Liq']=H2O_Liq
        if print is True:
            print('Water content replaced with that from H2O_Liq')

# Keiths Liq-only spreadsheet doesn't use Cr2O3 and P2O5 to calc cat. frac. So have set this to zero.
    if equationT != "T_Put2008_eq26_plag_sat" and equationT != "T_Put2008_eq24c_kspar_sat" \
    and equationT !="T_Put2005_eqD_plag_sat" and equationT != "T_Molina2015_amp_sat" and equationT != "T_Put2016_eq3_amp_sat":
        liq_comps_c['Cr2O3_Liq']=0
        liq_comps_c['P2O5_Liq']=0





# Now calculate cation fractions - if using putirka, uses hydrous cat fracs.
    if equationT == "T_Put2016_eq3_amp_sat":
        anhyd_cat_frac = calculate_hydrous_mol_fractions_liquid(liq_comps=liq_comps_c)
    else:
        anhyd_cat_frac = calculate_anhydrous_cat_fractions_liquid(liq_comps=liq_comps_c)

    if equationT=="T_Sug2000_eq6a_H7a" or equationT=="T_Sug2000_eq6b_H7b":
        anhyd_cat_frac['H2O_mol_frac']=calculate_hydrous_mol_fractions_liquid(liq_comps=liq_comps_c).H2O_Liq_mol_frac_hyd



# This performs extra calculation steps for Beattie equations
    if equationT == "T_Beatt93_BeattDMg" or equationT == "T_Beatt93_BeattDMg_HerzCorr" or equationT == "T_Put2008_eq19_BeattDMg" \
    or equationT == "T_Put2008_eq21_BeattDMg" or equationT == "T_Put2008_eq22_BeattDMg":
        anhyd_cat_frac['calcDMg_Beat93'] = (0.666 - (-0.049 * anhyd_cat_frac['Mn_Liq_cat_frac']
        + 0.027 * anhyd_cat_frac['Fet_Liq_cat_frac'])) / (
        1 * anhyd_cat_frac['Mg_Liq_cat_frac'] + 0.259 * anhyd_cat_frac['Mn_Liq_cat_frac']
        + 0.299 * anhyd_cat_frac['Fet_Liq_cat_frac'])

        anhyd_cat_frac['Beat_CNML'] = (anhyd_cat_frac['Mg_Liq_cat_frac'] + anhyd_cat_frac['Fet_Liq_cat_frac'] +
            anhyd_cat_frac['Ca_Liq_cat_frac'] + anhyd_cat_frac['Mn_Liq_cat_frac'])
        anhyd_cat_frac['Beat_CSiO2L'] = anhyd_cat_frac['Si_Liq_cat_frac']
        anhyd_cat_frac['Beat_NF'] = ( 7 / 2) * np.log(1 - anhyd_cat_frac['Al_Liq_cat_frac']) + 7 * np.log(1 - anhyd_cat_frac['Ti_Liq_cat_frac'])
        anhyd_cat_frac['Den_Beat93'] = 52.05 / 8.3144 + 2 * np.log(anhyd_cat_frac['calcDMg_Beat93']) + 2 * np.log(
            1.5 * anhyd_cat_frac['Beat_CNML']) + 2 * np.log(3 * anhyd_cat_frac['Beat_CSiO2L']) - anhyd_cat_frac['Beat_NF']

# Checks if P-dependent function you have entered a P
    if sig.parameters['P'].default is not None:
        if P is None:
            raise ValueError(f'{equationT} requires you to enter P, or specify P="Solve"')

    #else:
        # print('gothere2')
        # if P is not None:
        #     print('gothere3')
        #     print('Youve selected a P-independent function')
        #     P=None
        #     print('got here')



    kwargs = {name: anhyd_cat_frac[name] for name, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}
    if isinstance(P, str) or P is None:
        if P == "Solve":
            T_K = partial(func, **kwargs)
        if P is None:
            T_K=func(**kwargs)

    else:
        T_K=func(P, **kwargs)

    return T_K



def calculate_liq_only_temp_all_eqs(liq_comps, P=None, H2O_Liq=None):
    """ This function calculates Liq-only temps using a wide range of popular equations in the literature. Happy to add more to this on request!

    Parameters
    -------

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    H2O_Liq: float, int, pandas.Series, optional
        If users don't specify, uses H2O_Liq from liq_comps,
        if specified overwrites this.

    P: float, int, pandas.Series, optional
        Pressure in kbar

    Returns
    -------
    Pandas dataframe with temperature in kelvin from the following equations as column headings.



        Equations from Putirka (2008) and older studies:

            | T_Put2008_eq13
            | T_Put2008_eq14
            | T_Put2008_eq15
            | T_Put2008_eq16
            | T_Put2008_eq34_cpx_sat
            | T_Put1999_cpx_sat
            * Following 3 thermometers are adaptations of olivine-liquid thermometers with  DMg calculated using Beattie 1993,
            This means you can use them without knowing an olivine composition. ocan be applied when you haven't measured an olivine composiiton.
            | T_Put2008_eq19_BeattDMg
            | T_Put2008_eq21_BeattDMg
            | T_Put2008_eq22_BeattDMg

        Equations from Sugawara (2000):

            | T_Sug2000_eq1
            | T_Sug2000_eq3_ol
            | T_Sug2000_eq3_opx
            | T_Sug2000_eq3_cpx
            | T_Sug2000_eq3_pig
            | T_Sug2000_eq6a
            | T_Sug2000_eq6b

        Equations from Helz and Thornber (1987):

            | T_Helz1987_MgO
            | T_Helz1987_CaO

        Equation from Montrieth 1995
           | T_Montierth1995_MgO

        Equation from Beattie (1993)
           | T_Beatt1993_opx

        Equation from Shea et al. (2022)
           | T_Shea2022_MgO


        Equation from Molina et al. (2015)

            | T_Molina2015_amp_sat

        Equations from Putirka et al. (2016).
            | T_Put2016_eq3_amp_sat (saturation surface of amphibole)




    """




    # List of equation names
    equation_names = [
        'T_Put2008_eq13', 'T_Put2008_eq14', 'T_Put2008_eq15', 'T_Put2008_eq16',
        'T_Put2008_eq34_cpx_sat', 'T_Put1999_cpx_sat',
        'T_Put2008_eq19_BeattDMg', 'T_Put2008_eq21_BeattDMg', 'T_Put2008_eq22_BeattDMg',
        'T_Sug2000_eq1', 'T_Sug2000_eq3_ol', 'T_Sug2000_eq3_opx', 'T_Sug2000_eq3_cpx',
        'T_Sug2000_eq3_pig', 'T_Sug2000_eq6a', 'T_Sug2000_eq6b',
        'T_Helz1987_MgO', 'T_Helz1987_CaO', 'T_Montierth1995_MgO', 'T_Shea2022_MgO',
        'T_Molina2015_amp_sat', 'T_Put2016_eq3_amp_sat'
    ]


    # Dictionary to store results
    results = {}

    # Loop through each equation name, calculate temperature, and store in results
    for eq_name in equation_names:
        results[eq_name] = calculate_liq_only_temp(
            liq_comps=liq_comps, H2O_Liq=H2O_Liq, P=P, equationT=eq_name
        )



    return pd.DataFrame(results)





























import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import inspect


def rasmussen(P_kbar):
    """ 4th degree fit to the supporting information of Rasmussen et al. 2022,
    overall best fit density vs. depth. Above 5.24 kbar, returns Nan"""
    P=P_kbar
    if P<5.2474630296099205:
        b=-0.0025915704129682504
        c=0.037299399171806996
        d=-0.21276925206828018
        e=4.273285349609203
        f=0.00430844920014402
        D= b*(P**4) + c*(P**3) + d*(P**2) + e*P + f
    else:
        D=np.nan
    return D

def hill_zucca(P_kbar):
    """ Parameterization of Hill and Zucca (1987),
    as given in Putirka (2017) Down the Crater Elements supplement
    """
    P=P_kbar

    D=-2.77*10**(-5) * (P**4) - 2.0*10**(-3)*(P**3) - 4.88*10**(-2)*P**2 + 3.6*P - 6.34*(10**(-2))

    return D

def ryan_lerner(P_kbar):
    """ Parameterization of Ryan 1987, actual equation from Lerner et al. 2021
    After 16.88 km (455 MPa), returns Nan
    """
    P=P_kbar*100
    if P<455.09090909:
        D=(4.578*10**(-8) *P**3) - (4.151*10**(-5) *P**2) + (4.652*10**(-2) *P)
    else:
        D=np.nan

    return D

def denlinger_lerner(P_kbar):
    """ Calculates depth for a given pressure using the Parameterization of Denlinger and Flinders (2022) with the addition of 200 kg/m3 at shallow depths, from Lerner et al. (2024).
    After 15 km returns NaN following Lerner

    Parameters
    -------------
    P_kbar: int, float, pd.series
        Pressure in kbar

    Returns
    -------------
    Depth in km (same datatype as input)


    """

    D=0.0684*P_kbar**3 - 0.5008*P_kbar**2 + 4.47*P_kbar
    if D>15:
        D=np.nan

    return D


def mavko_debari(P_kbar):
    """ Parameterization of Mavko and Thompson (1983) and DeBari and Greene (2011)
    as given in Putirka (2017) Down the Crater Elements supplement, used for Cascades
    """
    P=P_kbar
    D=0.4853881 + 3.6006116*P - 0.0117368*(P-1.3822)**2


    return D


def prezzi(P_kbar):
    """ Parameterization of Prezzi et al. (2009),
    as given in Putirka (2017) Down the Crater Elements supplement.
    Used for Andes.
    """
    P=P_kbar
    D=4.88 + 3.30*P - 0.0137*(P - 18.01)**2

    return D


Profile_funcs={ryan_lerner, mavko_debari, hill_zucca, prezzi, rasmussen, denlinger_lerner}
Profile_funcs_by_name= {p.__name__: p for p in Profile_funcs}


## Two and three step profiles
def convert_pressure_depth_2step(P_kbar=None, d1=None, rho1=None, rho2=None, g=9.81):
    """ Converts Pressure to depth using a 2 step profile for int or float

    Parameters
    --------------
    P_kbar: int or float
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density

    rho1: int or float
        Density (kg/m3) down to step transition

    rho2: int or float
        Density (kg/m3) below step transition

    Returns
    -------------
    float
        Depth in km

    """

    d1_SI=d1*1000
    P_step1=(g*rho1*d1_SI)/100000000
    # print('Pressure Moho in kbar')
    # print(P_Moho)
    if P_kbar<P_step1:
        depth_km=10**(-3)*((P_kbar*100000000))/(g*rho1)
    if P_kbar>=P_step1:
        P_belowstep1=P_kbar-P_step1
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowstep1*100000000)/(g*rho2))
        depth_km=d1+depth_km_bm

    return depth_km

def loop_pressure_depth_2step(P_kbar=None, d1=14, rho1=2800, rho2=3100, g=9.81):

    """ Converts Pressure to depth using a 2 step profile for a pandas.Series of presssures

    Parameters
    --------------
    P_kbar: pd.Series
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density

    rho1: int or float
        Density (kg/m3) down to step transition

    rho2: int or float
        Density (kg/m3) below step transition

    Returns
    -------------
    pd.Series
        Depth in km

    """
    if np.isscalar(P_kbar):  # Check if it's a scalar (int, float, or numpy scalar)
        depth_km_loop=convert_pressure_depth_2step(P_kbar,
            d1=d1, rho1=rho1, rho2=rho2, g=g)
    else:
        depth_km_loop=np.empty(len(P_kbar))
        for i in range(0, len(P_kbar)):
            depth_km_loop[i]=convert_pressure_depth_2step(P_kbar[i],
            d1=d1, rho1=rho1, rho2=rho2, g=g)
    return depth_km_loop


def convert_pressure_depth_3step(P_kbar=None, d1=5, d2=14,
                                 rho1=2700, rho2=3000, rho3=3100, g=9.81):
    """ Converts Pressure to depth using a 3 step profile for int or float

    Parameters
    --------------
    P_kbar: int or float
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density

    d2: int or float
        depth in km of 2nd step transition in density

    rho1: int or float
        Density (kg/m3) down to first step transition

    rho2: int or float
        Density (kg/m3) between first and second step transition


    rho3: int or float
        Density (kg/m3) below 2nd step transition

    Returns
    -------------
    float
        Depth in km

    """

    d1_SI=d1*1000
    d2_SI=d2*1000
    P_Step1=(g*rho1*d1_SI)/100000000
    P_Step2=P_Step1+(g*(d2_SI-d1_SI)*rho2)/100000000
    # print('Pressure Moho in kbar')
    # print(P_Moho)
    if P_kbar<P_Step1:
        depth_km=10**(-3)*((P_kbar*100000000))/(g*rho1)
    if P_kbar>=P_Step1 and P_kbar<P_Step2:
        P_belowStep2=P_kbar-P_Step1
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowStep2*100000000)/(g*rho2))
        depth_km=d1+depth_km_bm
    if P_kbar>=P_Step2:
        P_belowstep2=P_kbar-P_Step2
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowstep2*100000000)/(g*rho3))
        depth_km=d2+depth_km_bm

    return depth_km




def loop_pressure_depth_3step(P_kbar=None,  d1=5, d2=14,
                                 rho1=2700, rho2=3000, rho3=3100, g=9.81):

    """ Converts Pressure to depth using a 3 step profile for pd.Series

    Parameters
    --------------
    P_kbar: pd.Series
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density

    d2: int or float
        depth in km of 2nd step transition in density

    rho1: int or float
        Density (kg/m3) down to first step transition

    rho2: int or float
        Density (kg/m3) between first and second step transition


    rho3: int or float
        Density (kg/m3) below 2nd step transition

    Returns
    -------------
    pd.Series
        Depth in km

    """
    if np.isscalar(P_kbar):  # Check if it's a scalar (int, float, or numpy scalar)
        depth_km_loop=convert_pressure_depth_3step(P_kbar=P_kbar,
            d1=d1, d2=d2,rho1=rho1, rho2=rho2, rho3=rho3, g=g)
    else:

        depth_km_loop=np.empty(len(P_kbar))
        for i in range(0, len(P_kbar)):
            depth_km_loop[i]=convert_pressure_depth_3step(P_kbar=P_kbar[i],
            d1=d1, d2=d2,rho1=rho1, rho2=rho2, rho3=rho3, g=g)
    return depth_km_loop

########## Same for 4 step

def convert_pressure_depth_4step(*, P_kbar=None, d1=None, d2=None, d3=None,
                                 rho1=None, rho2=None, rho3=None, rho4=None, g=9.81):
    """ Converts Pressure to depth using a 3 step profile for int or float

    Parameters
    --------------
    P_kbar: int or float
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density from the surface

    d2: int or float
        depth in km of 2nd step transition in density from the surface

    d3: int or float
        depth in km of 3rd step transition in density from the surface




    rho1: int or float
        Density (kg/m3) down to first step transition

    rho2: int or float
        Density (kg/m3) between first and second step transition

    rho3: int or float
        Density (kg/m3) below 2nd step transition

    rho4: int or float
        Density (kg/m3) below 3rd step transition

    Returns
    -------------
    float
        Depth in km

    """

    d1_SI=d1*1000
    d2_SI=d2*1000
    d3_SI=d3*1000
    P_Step1=(g*rho1*d1_SI)/100000000 # Convert to kbar
    P_Step2=P_Step1+(g*(d2_SI-d1_SI)*rho2)/100000000 # Pressure contribution from step 2
    P_Step3=P_Step2+(g*(d3_SI-d2_SI)*rho3)/100000000

    # If the pressure is less than Step 1, easy math, just use Rho1
    if P_kbar<P_Step1:
        depth_km=10**(-3)*((P_kbar*100000000))/(g*rho1)

    # If the pressure is between step 1 and step 2, First, take away the pressure of step 1, then solve for height in layer 2
    if P_kbar>=P_Step1 and P_kbar<P_Step2:
        P_belowStep2=P_kbar-P_Step1
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowStep2*100000000)/(g*rho2))
        depth_km=d1+depth_km_bm

    # If the pressure is between step 2 and 3, take away the pressure of step2, and calculate the height within layer 3
    if P_kbar>=P_Step2 and P_kbar<P_Step3:
        P_belowstep2=P_kbar-P_Step2
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowstep2*100000000)/(g*rho3))
        depth_km=d2+depth_km_bm

    # If the pressure is greater than that at the base of layer 3, take away the pressure of step3, and calculate the height within layer 4
    if P_kbar>=P_Step3:
        P_belowstep3=P_kbar-P_Step3
        # print('P below  Moho')
        # print(P_belowMoho)
        depth_km_bm=10**(-3)*((P_belowstep3*100000000)/(g*rho4))
        depth_km=d3+depth_km_bm

    return depth_km

def loop_pressure_depth_4step(*, P_kbar=None,  d1=5, d2=14, d3=20,
                                 rho1=2700, rho2=3000, rho3=3100, rho4=4000, g=9.81):

    """ Converts Pressure to depth using a 3 step profile for pd.Series

    Parameters
    --------------
    P_kbar: pd.Series
        Pressure in kbar

    d1: int or float
        depth in km of 1st step transition in density

    d2: int or float
        depth in km of 2nd step transition in density

    rho1: int or float
        Density (kg/m3) down to first step transition

    rho2: int or float
        Density (kg/m3) between first and second step transition


    rho3: int or float
        Density (kg/m3) below 2nd step transition

    Returns
    -------------
    pd.Series
        Depth in km

    """
    if np.isscalar(P_kbar):  # Check if it's a scalar (int, float, or numpy scalar)
        depth_km_loop=convert_pressure_depth_4step(P_kbar=P_kbar,
            d1=d1, d2=d2, d3=d3, rho1=rho1, rho2=rho2, rho3=rho3, rho4=rho4, g=g)
    else:

        depth_km_loop=np.empty(len(P_kbar))
        for i in range(0, len(P_kbar)):
            depth_km_loop[i]=convert_pressure_depth_4step(P_kbar=P_kbar[i],
            d1=d1, d2=d2, d3=d3, rho1=rho1, rho2=rho2, rho3=rho3, rho4=rho4, g=g)
    return depth_km_loop

#######3


def convert_pressure_to_depth(*, P_kbar=None, crust_dens_kgm3=None, g=9.81,
d1=None, d2=None, d3=None, rho1=None, rho2=None, rho3=None, rho4=None, model=None):
    """ Converts pressure in kbar to depth in km using a variety of crustal density profiles


    Parameters
    -----------

    P_kbar: int, float, pd.Series, np.ndarray
        Pressure in kbar

    g: float
        gravitational constant, in m/s2

    crust_dens_kgm3: float or str
        If float: Crustal density in kg/m3

        If model, choose from:

        ryan_lerner:
            Parameterization of Ryan 1987, actual equation from Lerner et al. 2021
            After 16.88 km (455 MPa), assume density is 2.746, as density turns around again. This profile is tweaked for Hawaii

        denlinger_lerner:
            Parameterization of Denlinger and Flinders (2022) with the addition of 200 kg/m3 at shallow depths, from Lerner et al. (2024).
            After 15 km returns NaN following Lerner

        mavko_debari:
            Parameterization of Mavko and Thompson (1983) and DeBari and Greene (2011)
            as given in Putirka (2017) Down the Crater Elements supplement.


        hill_zucca:
            Parameterization of Hill and Zucca (1987),
            as given in Putirka (2017) Down the Crater Elements supplement

        prezzi:
            Parameterization of Prezzi et al. (2009),
            as given in Putirka (2017) Down the Crater Elements supplement. Tweaked for Andes.

        rasmussen:
            Linear fit to the supporting information of Rasmussen et al. 2022,
            overall best fit density vs. depth

        two-step:
            If two step, must also define d1 (depth to 1st step), rho1 (density to 1st step), rho2 (density to 2nd step)

        three-step:
            If three step, must also define d1 (depth to 1st step), d2 (depth to second step), rho1 (density to 1st step), rho2 (density to 2nd step),
            rho3 (density after 3rd step) in km and kg/m3 respectively.


        four-step:
            If four step, must also define d1 (depth to 1st step from surface), d2 (depth to second step from surface), d3 (depth to 3rd step from surface)  rho1 (density to 1st step), rho2 (density to 2nd step),
            rho3 (density to 3rd step), and rho4 (density below 3rd step) in km and kg/m3 respectively.

    OR



    Else, just enter a crustal density in kg/m3, e.g., model=2700



    Returns
    -----------

    Depth in km as a panda series

    """


    # Check, is it an integer, If so just calculate depth
    if type(crust_dens_kgm3)==int or type(crust_dens_kgm3)==float:
        D=10**5*P_kbar/(g*crust_dens_kgm3)
        model=None


    else:
        # Check if its a pandas series.
        if isinstance(crust_dens_kgm3, pd.Series):
            # Now check, is it a series of strings, or
            if type(crust_dens_kgm3[0])==str:
                model=crust_dens_kgm3.iloc[0]
            else:
                model=None
                D=10**5*P_kbar/(g*crust_dens_kgm3)

        # Check if its just a single string
        elif type(crust_dens_kgm3)==str:
            model=crust_dens_kgm3



    if model is not None:
        if model == "two-step":
            if d1 is None or rho1 is None or rho2 is None:
                raise Exception('You have selected the two-step model, You must enter d1 (km), rho1 and rho2 (kg/m3)')
            D=loop_pressure_depth_2step(P_kbar=P_kbar,
            d1=d1, rho1=rho1, rho2=rho2, g=g)

        if model == "three-step":
            if d1 is None or d2 is None or rho1 is None or rho2 is None or rho3 is None:
                raise Exception('You have selected the two-step model, You must enter d1 and d2 (km), rho1, rho2 and rho3 (kg/m3)')

            D=loop_pressure_depth_3step(P_kbar=P_kbar,
            d1=d1, d2=d2, rho1=rho1, rho2=rho2, rho3=rho3, g=g)

        if model == "four-step":
            if d1 is None or d2 is None or d3 is None or rho1 is None or rho2 is None or rho3 is None or rho4 is None:
                raise Exception('You have selected the two-step model, You must enter d1 and d2 (km), rho1, rho2 and rho3 (kg/m3)')

            D=loop_pressure_depth_4step(P_kbar=P_kbar,
            d1=d1, d2=d2, d3=d3, rho1=rho1, rho2=rho2, rho3=rho3, rho4=rho4, g=g)

        if model !="two-step" and model != "three-step" and model != "four-step":
            try:

                func = Profile_funcs_by_name[model]
            except KeyError:
                raise ValueError(f'{model} is not a valid model') from None

            sig=inspect.signature(func)

            if isinstance(P_kbar, float) or isinstance(P_kbar, int):
                D=func(P_kbar)

            if isinstance(P_kbar, pd.Series):
                D=np.empty(len(P_kbar), float)
                for i in range(0, len(P_kbar)):
                    D[i]=func(P_kbar.iloc[i])

            if isinstance(P_kbar, np.ndarray):
                D=np.empty(len(P_kbar), float)
                for i in range(0, len(P_kbar)):
                    D[i]=func(P_kbar[i])




    D_series=pd.Series(D)



    return D_series




import numpy as np
import matplotlib.pyplot as plt
from functools import partial
import inspect
import warnings as w
import numbers
import pandas as pd


from Thermobar.core import *
from Thermobar.mineral_equilibrium import *

## Functions for olivine-liquid hygrometry
def H_Gavr2016(*, CaO_Liq, MgO_Liq, CaO_Ol):
    '''
    Olivine-Liquid hygrometer of Gavrilenko et al. (2016).
    '''
    DCa_dry_HighMgO=0.00042*MgO_Liq+0.0196
    DCa_dry_LowMgO=-0.0043*MgO_Liq+0.072
    DCa_dry_calc=np.empty(len(MgO_Liq), dtype=float)
    DeltaCa=np.empty(len(MgO_Liq), dtype=float)

    H2O_Calc=np.empty(len(MgO_Liq), dtype=float)

    DCa_Meas=CaO_Ol/CaO_Liq
    DCa_Divider=0.00462*MgO_Liq-0.027
    for i in range(0, len(MgO_Liq)):
        if DCa_Meas[i]<=DCa_Divider[i]:
            DCa_dry_calc[i]=DCa_dry_HighMgO[i]
            DeltaCa[i]=DCa_dry_calc[i]-DCa_Meas[i]
            H2O_Calc[i]=397*DeltaCa[i]
        else:
            DCa_dry_calc[i]=DCa_dry_LowMgO[i]
            DeltaCa[i]=DCa_dry_calc[i]-DCa_Meas[i]
            H2O_Calc[i]=188*DeltaCa[i]
    H2O_Calc[CaO_Ol==0]=np.nan
    return H2O_Calc


def H_Pu2017(*, NiO_Ol_mol_frac, FeOt_Liq_mol_frac, MnO_Liq_mol_frac, MgO_Liq_mol_frac,
             CaO_Liq_mol_frac, NiO_Liq_mol_frac, Al2O3_Liq_mol_frac, TiO2_Liq_mol_frac, SiO2_Liq_mol_frac,MgO_Ol_mol_frac):

    T_Pu2017_calc = T_Pu2017(
        NiO_Ol_mol_frac=NiO_Ol_mol_frac,
        FeOt_Liq_mol_frac=FeOt_Liq_mol_frac,
        MnO_Liq_mol_frac=MnO_Liq_mol_frac,
        MgO_Liq_mol_frac=MgO_Liq_mol_frac,
        CaO_Liq_mol_frac=CaO_Liq_mol_frac,
        NiO_Liq_mol_frac=NiO_Liq_mol_frac,
        Al2O3_Liq_mol_frac=Al2O3_Liq_mol_frac,
        TiO2_Liq_mol_frac=TiO2_Liq_mol_frac,
        SiO2_Liq_mol_frac=SiO2_Liq_mol_frac
    )


    T_Pu2017_Mg_calc = T_Pu2017_Mg(
        MgO_Ol_mol_frac=MgO_Ol_mol_frac,
        FeOt_Liq_mol_frac=FeOt_Liq_mol_frac,
        MnO_Liq_mol_frac=MnO_Liq_mol_frac,
        NiO_Liq_mol_frac=NiO_Liq_mol_frac,
        CaO_Liq_mol_frac=CaO_Liq_mol_frac,
        MgO_Liq_mol_frac=MgO_Liq_mol_frac,
        Al2O3_Liq_mol_frac=Al2O3_Liq_mol_frac,
        TiO2_Liq_mol_frac=TiO2_Liq_mol_frac,
        SiO2_Liq_mol_frac=SiO2_Liq_mol_frac
    )



    deltaT=T_Pu2017_Mg_calc-T_Pu2017_calc
    MinH2O=0.000069171*(deltaT**2)+0.02615*deltaT

    return MinH2O



Liquid_olivine_hygr_funcs = {H_Gavr2016, H_Pu2017}
Liquid_olivine_hygr_funcs_by_name = {p.__name__: p for p in Liquid_olivine_hygr_funcs}


def calculate_ol_liq_hygr(*, liq_comps=None, ol_comps=None, equationH=None, eq_tests=False,
P=None, T=None, meltmatch=None, equationT=None, Fe3Fet_Liq=None):
    '''
    Olivine-liquid hygrometer. Returns the estimated H2O content
    in wt%

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    ol_comps: pandas.DataFrame
        Olivine compositions with column headings SiO2_Ol, MgO_Ol etc.

    Or

    meltmatch: prematched olivines and liquids if you are using a matching function.

    equationH: str
        H_Gavr2016 (P-independent, H2O_independent)
        H_Pu2017 (uses the difference between Tmg and Tni to calculate a minimum H2O content. requires NiO in liquid and olivine to be entered).


    eq_tests: bool
        if true, calculates Kd for olivine-liquid pairs.
        Other inputs for these tests:

        P: int, flt, pandas.Series (needed for Toplis Kd calculation).
        If nothing inputted, P set to 1 kbar

        T: int, flt, pandas.Series (needed for Toplis KD calculation).
        Can also specify equationT="" to calculate temperature
        using an olivine-liquid thermometer, using the calculated H2O content
        from the hygrometer

        Fe3Fet_Liq: int, flt, pandas.Series. As Kd calculated using only Fe2 in the Liq.


    Returns
    -------
    pandas.core.series.Series
        H2O content in wt%.
 '''
# These are checks that our inputs are okay
    if meltmatch is None:
        if len(liq_comps)!=len(ol_comps):
            raise ValueError('Ol comps need to be same length as Liq comps. use a _matching function calculate_ol_liq_hygr_matching instead if you want to consider all pairs')



    try:
        func = Liquid_olivine_hygr_funcs_by_name[equationH]
    except KeyError:
        raise ValueError(f'{equationH} is not a valid equation') from None
    sig=inspect.signature(func)

    if meltmatch is None:
        ol_comps_c=ol_comps.copy()
        liq_comps_c=liq_comps.copy()

        if equationH == "H_Pu2017":

            anhyd_mol_frac_Ni = calculate_anhydrous_mol_fractions_liquid_Ni(
                liq_comps=liq_comps_c)
            # Calculate the liquid mole fraction of NiO
            ol_mol_frac_Ni = calculate_mol_fractions_olivine_ni(
                    ol_comps=ol_comps_c)

            Liq_Ols = pd.concat([anhyd_mol_frac_Ni, ol_mol_frac_Ni, ol_comps_c, liq_comps_c], axis=1)

            Liq_Ols = pd.concat([anhyd_mol_frac_Ni, ol_mol_frac_Ni, ol_comps_c, liq_comps_c], axis=1)



        else:
            if Fe3Fet_Liq is not None:
                liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq

            if len(liq_comps) != len(ol_comps):
                raise ValueError('The panda series entered for olivine isnt the same length as for liquids')

            anhyd_cat_frac = calculate_anhydrous_cat_fractions_liquid(liq_comps=liq_comps_c)
            ol_cat_frac = calculate_cat_fractions_olivine(ol_comps=ol_comps_c)
            Liq_Ols = pd.concat([ol_comps_c, anhyd_cat_frac, ol_cat_frac], axis=1)

    if meltmatch is not None:


        Liq_Ols=meltmatch






    kwargs = {name: Liq_Ols[name] for name, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}

    H2O_Calc_np=func(**kwargs)
    H2O_Calc=pd.Series(H2O_Calc_np)



    if eq_tests is False and meltmatch is None:
         return H2O_Calc

    if eq_tests is False and meltmatch is not None:
        if 'H2O_calc' in Liq_Ols.columns:
            print("H2O_Calc Column already exists in dataframe. Have ovewritten")
            Liq_Ols['H2O_calc']=H2O_Calc

        else:
            Liq_Ols.insert(0, 'H2O_calc', H2O_Calc)

        return Liq_Ols

    if eq_tests is True:
        if P is None:
            P = 1
            print(
                'You have not selected a pressure, so we have calculated Toplis Kd at 1kbar')
        if T is None and equationT is None:
            raise ValueError('Temperature needed to calculate Delta Kd using Toplis.'
             'Either enter T, or specify equationT=""')

        if equationT is not None:
            # If not melt match, do separatly
            #if meltmatch is None:

            T=calculate_ol_liq_temp(meltmatch=Liq_Ols,
            equationT=equationT, H2O_Liq=H2O_Calc, P=P, eq_tests=eq_tests).T_K_calc
            # If melt match do on combined dataframe
            # if meltmatch is not None:
            #     T=calculate_ol_liq_temp(ol_comps=ol_comps, liq_comps=liq_comps,
            # equationT=equationT, H2O_Liq=H2O_Calc, P=P).T_K_calc

        ol_fo = (Liq_Ols['MgO_Ol'] / 40.3044) / \
            ((Liq_Ols['MgO_Ol'] / 40.3044) + Liq_Ols['FeOt_Ol'] / 71.844)
        KdFeMg_Meas = (
            ((Liq_Ols['FeOt_Ol'] / 71.844) / (Liq_Ols['MgO_Ol'] / 40.3044)) /
            ((Liq_Ols['FeOt_Liq'] * (1 - Liq_Ols['Fe3Fet_Liq']
                                         ) / 71.844) / (Liq_Ols['MgO_Liq'] / 40.3044))
        )
        Kd_func = partial(calculate_toplis2005_kd, SiO2_mol=Liq_Ols['SiO2_Liq_mol_frac'],
        Na2O_mol=Liq_Ols['Na2O_Liq_mol_frac'], K2O_mol=Liq_Ols['Na2O_Liq_mol_frac'],
        P=P, H2O=Liq_Ols['H2O_Liq'], T=T)
        Kd_Toplis_calc = Kd_func(ol_fo)

        DeltaKd_Toplis = KdFeMg_Meas - Kd_Toplis_calc
        DeltaKd_Roeder = KdFeMg_Meas - 0.3
        DeltaKd_Matzen = KdFeMg_Meas - 0.34
        if equationT is None:
            df = pd.DataFrame(data={'H2O_calc': H2O_Calc, 'Temp used for calcs': T, 'P used for calcs': P, 'Kd Meas': KdFeMg_Meas, 'Kd calc (Toplis)': Kd_Toplis_calc,'ΔKd, Toplis (M-P)': DeltaKd_Toplis, 'ΔKd, Roeder (M-P)': DeltaKd_Roeder, 'ΔKd, Matzen (M-P)': DeltaKd_Matzen})
        if equationT is not None:
            df = pd.DataFrame(data={'H2O_calc': H2O_Calc, 'T_K_calc': T, 'P used for calcs': P, 'Kd Meas': KdFeMg_Meas, 'Kd calc (Toplis)': Kd_Toplis_calc,'ΔKd, Toplis (M-P)': DeltaKd_Toplis, 'ΔKd, Roeder (M-P)': DeltaKd_Roeder, 'ΔKd, Matzen (M-P)': DeltaKd_Matzen})


        df_out = pd.concat([df, Liq_Ols], axis=1)

        return df_out

def calculate_ol_liq_hygr_matching(*, liq_comps, ol_comps, equationH, eq_tests=False,
T=None, equationT=None, P=None,  Fe3Fet_Liq=None, iterations=30):

    '''
    Evaluates all possible Ol-liq pairs for H2O
    returns H2O and equilibrium test values.

    Parameters
    ----------------

    ol_comps: pandas.DataFrame
        Panda DataFrame of Ol compositions with column headings SiO2_Ol, CaO_Ol etc.

    liq_comps: pandas.DataFrame
        Panda DataFrame of liq compositions with column headings SiO2_Liq etc.

    EquationH: str
        Choose from:

            |   H_Gavr2016 (P-independent, T-independent)


    iterations: int
        number of times to iterate temperature and H2O. Default 30.

    Returns
    -------
    pandas.DataFrame
        H2O (wt%) for all posible ol-liq matches, along with equilibrium
        tests, components and input mineral compositions
    '''

    ol_comps_c=ol_comps.copy()
    liq_comps_c=liq_comps.copy()

    #Check valid equation for T
    try:
        func = Liquid_olivine_hygr_funcs_by_name[equationH]
    except KeyError:
        raise ValueError(f'{equationH} is not a valid equation for Ol-Liquid Hygrometry') from None
    sig=inspect.signature(func)

    # This is a duplication step, to pair up all possible mathes. First we give a unique identifier

    ol_comps_c['ID_Ol'] = ol_comps_c.index
    liq_comps_c['ID_liq'] = liq_comps_c.index.astype('float64')

    if "Sample_ID_Ol" not in ol_comps:
        ol_comps_c['Sample_ID_Ol'] = ol_comps.index.astype('float64')
    if "Sample_ID_liq" not in liq_comps:
        liq_comps_c['Sample_ID_liq'] = liq_comps.index.astype('float64')


    # This calculates components proir to duplication (more computationally efficient)

    if equationH !='H_Pu2017':
        anhyd_liq_frac = calculate_anhydrous_cat_fractions_liquid(liq_comps=liq_comps_c)
        ol_cat_frac1 = calculate_cat_fractions_olivine(ol_comps=ol_comps_c)
        ol_cat_frac= pd.concat([ol_comps_c, ol_cat_frac1], axis=1)

    if equationH=='H_Pu2017':

            anhyd_mol_frac_Ni = calculate_anhydrous_mol_fractions_liquid_Ni(
                liq_comps=liq_comps_c)
            # Calculate the liquid mole fraction of NiO
            ol_mol_frac_Ni = calculate_mol_fractions_olivine_ni(
                    ol_comps=ol_comps_c)

            ol_cat_frac = pd.concat([ ol_mol_frac_Ni, ol_comps_c], axis=1)
            anhyd_liq_frac=pd.concat([anhyd_mol_frac_Ni, liq_comps_c])





    # This duplicates Ols, repeats liq1-liq1*N, liq2-liq2*N etc.
    DupOls = pd.DataFrame(np.repeat(ol_cat_frac.values, np.shape(
        anhyd_liq_frac)[0], axis=0))  # .astype('float64')
    DupOls.columns = ol_cat_frac.columns

    # This duplicates liquids like liq1-liq2-liq3 for liq1, liq1-liq2-liq3 for
    # liq2 etc.
    Dupliqs = pd.concat([anhyd_liq_frac ] * np.shape(ol_cat_frac)[0]).reset_index(drop=True)
    # Combines these merged liquids and liq dataframes
    Combo_Ol_liqs = pd.concat([Dupliqs, DupOls], axis=1)

    Combo_Ol_liqs_1 = Combo_Ol_liqs.copy()
    # Combo_Ols_liqs_1['K_Barth'] = Combo_Ols_liqs_1['Ab_liq'] / \
    #     Combo_Ols_liqs_1['Ab_Ol']
    LenCombo = str(np.shape(Combo_Ol_liqs)[0])

    LenOl=len(ol_comps)
    LenLiqs=len(liq_comps)
    print("Considering N=" + str(LenOl) + " Ol & N=" + str(LenLiqs) +" Liqs, which is a total of N="+ str(LenCombo) +
          " Liq- Ol pairs, be patient if this is >>1 million!")





    CalcH2O=calculate_ol_liq_hygr(meltmatch=Combo_Ol_liqs_1,
    equationH=equationH, T=T, P=P, eq_tests=eq_tests,
    Fe3Fet_Liq=Fe3Fet_Liq)


    return CalcH2O





## Functions for olivine-Liquid thermometry
# Defining functions where DMg is measured.
def T_Beatt93_ol(P, *, Den_Beat93):
    '''
    Olivine-Liquid thermometer: Beattie (1993). Putirka (2008) suggest this is the best olivine-liquid thermometer for anhydrous conditions at relatively low pressures.
    '''
    return (((113.1 * 1000) / 8.3144 + (0.1 * P * 10**9 - 10**5)
            * 4.11 * (10**(-6)) / 8.3144) / Den_Beat93)
# This is equation 19 from Putirka 2008


def T_Beatt93_ol_HerzCorr(P, *, Den_Beat93):
    '''
    Olivine-Liquid thermometer: Beattie (1993) with correction of Herzberg and O'Hara (2002), eliminating systematic error at higher pressures
    Anhydrous SEE=±44°C
    Hydrous SEE=±53°C
    '''

    return (((113.1 * 1000 / 8.3144 + (0.0001 * (10**9) - 10**5) * 4.11 *
            (10**(-6)) / 8.3144) / Den_Beat93) + 54 * (0.1 * P) + 2 * (0.1 * P)**2)


def T_Put2008_eq19(P, *, DMg_Meas, CNML, CSiO2L, NF):
    '''
    Olivine-Liquid thermometer originally from Beattie, (1993), form from Putirka (2008)

    '''
    return ((13603) + (4.943 * 10**(-7)) * ((0.1 * P)*10**9 - 10**(-5))) / (6.26 + 2 *
            np.log(DMg_Meas) + 2 * np.log(1.5 * CNML) + 2 * np.log(3 * CSiO2L) - NF)



def T_Put2008_eq21(P, *, DMg_Meas, Na2O_Liq, K2O_Liq, H2O_Liq):
    '''
    Olivine-Liquid thermometer: Putirka (2008), equation 21 (originally Putirka et al., 2007,  Eq 2). Recalibration of Beattie (1993) to account for the pressures sensitivity noted by Herzberg and O'Hara (2002), and esliminates the systematic error of Beattie (1993) for hydrous compositions.
    Anhydrous SEE=±53°C
    Hydrous SEE=±36°C
    '''
    return (1 / ((np.log(DMg_Meas) + 2.158 - 5.115 * 10**(-2) * (Na2O_Liq +
            K2O_Liq) + 6.213 * 10**(-2) * H2O_Liq) / (55.09 * 0.1 * P + 4430)) + 273.15)


def T_Put2008_eq22(P, *, DMg_Meas, CNML, CSiO2L, NF, H2O_Liq):
    '''
    Olivine-Liquid thermometer: Putirka (2008), equation 22 (originally Putirka et al., 2007,  Eq 4). Recalibration of Beattie (1993) to account for the pressures sensitivity noted by Herzberg and O'Hara (2002), and esliminates the systematic error of Beattie (1993) for hydrous compositions. Putirka (2008) suggest this is the best olivine-liquid thermometer for hydrous melts.
    Anhydrous SEE±45°C
    Hydrous SEE±29°C
    '''
    return ((15294.6 + 1318.8 * 0.1 * P + 2.48348 * ((0.1 * P)**2)) / (8.048 + 2.8352 * np.log(DMg_Meas) + 2.097 *
            np.log(1.5 * CNML) + 2.575 * np.log(3 * CSiO2L) - 1.41 * NF + 0.222 * H2O_Liq + 0.5 * (0.1 * P)) + 273.15)


def T_Sisson1992(P, *, KdMg_TSG1992):
    '''
    Olivine-Liquid thermometer: Sisson and Grove (1992). Putirka (2008) suggests this thermometer is best in peridotitic systems containing 2-25 wt% CO2.
    '''
    return ((4129 + 0.0146 * (P*1000-1)) /
            (np.log10(KdMg_TSG1992) + 2.082))


def T_Pu2017(P=None, *, NiO_Ol_mol_frac, FeOt_Liq_mol_frac, MnO_Liq_mol_frac, MgO_Liq_mol_frac,
             CaO_Liq_mol_frac, NiO_Liq_mol_frac, Al2O3_Liq_mol_frac, TiO2_Liq_mol_frac, SiO2_Liq_mol_frac):
    '''
    Olivine-Liquid thermometer: Pu et al. (2017), Eq 1  Uses D Ni (ol-melt) rather than D Mg (ol-melt),
    meaning this thermometer has far less sensitivity to H2O or pressure at 0-1 GPa.
    SEE=±29°C
    '''
    # Check it has Ni
    NiO_Ol_mol_frac = np.asarray(NiO_Ol_mol_frac, dtype=float)
    NiO_Liq_mol_frac = np.asarray(NiO_Liq_mol_frac, dtype=float)

    hasNi=(NiO_Ol_mol_frac>0) & (NiO_Liq_mol_frac>0)

    D_Ni_Mol = np.divide(NiO_Ol_mol_frac, NiO_Liq_mol_frac, out=np.full_like(NiO_Ol_mol_frac, np.nan), where=hasNi)
    XNm = FeOt_Liq_mol_frac + MnO_Liq_mol_frac + \
        MgO_Liq_mol_frac + CaO_Liq_mol_frac + NiO_Liq_mol_frac
    NFX = 3.5 * np.log(1 - Al2O3_Liq_mol_frac) + 7 * \
        np.log(1 - TiO2_Liq_mol_frac)


 # Calculate the log terms only for valid entries
    valid_indices = (D_Ni_Mol > 0) & (XNm > 0) & (SiO2_Liq_mol_frac > 0)

    # Calculate the temperature for valid values

    temperature = np.where(valid_indices,
                       9416 / (np.log(D_Ni_Mol.astype(float)) + 0.71 * np.log(XNm.astype(float)) -
                                0.349 * NFX - 0.532 * np.log(SiO2_Liq_mol_frac.astype(float)) + 4.319),
                       np.nan)



    return temperature

def T_Pu2017_Mg(P=None, *, MgO_Ol_mol_frac, FeOt_Liq_mol_frac, MnO_Liq_mol_frac,NiO_Liq_mol_frac,
             CaO_Liq_mol_frac, MgO_Liq_mol_frac, Al2O3_Liq_mol_frac, TiO2_Liq_mol_frac, SiO2_Liq_mol_frac):
    '''
    Olivine-Liquid thermometer: Pu et al. (2017), Eq 1  Uses D Mg (ol-melt)
    '''
    # Check it has Ni
    MgO_Ol_mol_frac = np.asarray(MgO_Ol_mol_frac, dtype=float)
    MgO_Liq_mol_frac = np.asarray(MgO_Liq_mol_frac, dtype=float)

    hasNi=(NiO_Liq_mol_frac>0)

    D_Mg_Mol = np.divide(MgO_Ol_mol_frac, MgO_Liq_mol_frac, out=np.full_like(MgO_Ol_mol_frac, np.nan), where=hasNi)
    XNm = FeOt_Liq_mol_frac + MnO_Liq_mol_frac + \
        MgO_Liq_mol_frac + CaO_Liq_mol_frac + NiO_Liq_mol_frac
    NFX = 3.5 * np.log(1 - Al2O3_Liq_mol_frac) + 7 * \
        np.log(1 - TiO2_Liq_mol_frac)


 # Calculate the log terms only for valid entries
    valid_indices = (D_Mg_Mol > 0) & (XNm > 0) & (SiO2_Liq_mol_frac > 0)

    # Calculate the temperature for valid values

    temperature = np.where(valid_indices, 6701/(np.log(D_Mg_Mol.astype(float))+1.12*np.log(XNm.astype(float))-0.64*NFX+1.08*np.log(SiO2_Liq_mol_frac.astype(float))+4.74), np.nan)


    return temperature


def T_Pu2021(P, *, NiO_Ol_mol_frac, FeOt_Liq_mol_frac, MnO_Liq_mol_frac, MgO_Liq_mol_frac,
             CaO_Liq_mol_frac, NiO_Liq_mol_frac, Al2O3_Liq_mol_frac, TiO2_Liq_mol_frac, SiO2_Liq_mol_frac):
    '''
    Olivine-Liquid thermometer: Pu et al. (2017), with the pressure correction of Pu et al. (2021).
    Uses D Ni (ol-melt) rather than D Mg (ol-melt), meaning this thermometer has far less sensitivity to melt H2O than other olivine-liquid thermometers.
    SEE=±45°C (for the 2017 expression).


    '''

    # Get P to the right format

    if isinstance(P, (float, int)):
            P_values = pd.Series([P] * len(NiO_Ol_mol_frac))
    elif isinstance(P, pd.Series):
        P_values = P
    else:
        raise ValueError("P must be either a float, integer, or a pandas Series.")

    NiO_Ol_mol_frac = np.asarray(NiO_Ol_mol_frac, dtype=float)
    NiO_Liq_mol_frac = np.asarray(NiO_Liq_mol_frac, dtype=float)

    hasNi=(NiO_Ol_mol_frac>0) & (NiO_Liq_mol_frac>0)

    D_Ni_Mol = np.divide(NiO_Ol_mol_frac, NiO_Liq_mol_frac, out=np.full_like(NiO_Ol_mol_frac, np.nan), where=hasNi)

    XNm = FeOt_Liq_mol_frac + MnO_Liq_mol_frac + \
        MgO_Liq_mol_frac + CaO_Liq_mol_frac + NiO_Liq_mol_frac
    NFX = 3.5 * np.log(1 - Al2O3_Liq_mol_frac) + 7 * \
        np.log(1 - TiO2_Liq_mol_frac)
# Initialize P_Corr to zero for all rows
    P_Corr = pd.Series(np.zeros(len(P_values)), index=P_values.index)

    # Calculate P_Corr only where P is between 10 and 30 GPa
    mask = (P_values >= 10) & (P_values <= 30)
    P_Corr[mask] = - 70 + 110 * (P_values[mask] * 0.1) - 18 * (P_values[mask] * 0.1)**2



 # Calculate the log terms only for valid entries
    valid_indices = (D_Ni_Mol > 0) & (XNm > 0) & (SiO2_Liq_mol_frac > 0)

    # Calculate the temperature for valid values
    temperature = np.where(valid_indices,
                       9416 / (np.log(D_Ni_Mol.astype(float)) + 0.71 * np.log(XNm.astype(float)) - 0.349 * NFX - 0.532 * np.log(SiO2_Liq_mol_frac.astype(float)) + 4.319),
                       np.nan)

    return temperature








## Listing all equation options
Liquid_olivine_funcs = {T_Beatt93_ol, T_Beatt93_ol_HerzCorr, T_Put2008_eq19, T_Put2008_eq21,
T_Put2008_eq22, T_Sisson1992, T_Pu2017, T_Pu2017_Mg, T_Pu2021}

Liquid_olivine_funcs_by_name = {p.__name__: p for p in Liquid_olivine_funcs}


## Function for calculating olivine-liquid temperature using various equations.

def calculate_ol_liq_temp_matching(*, liq_comps, ol_comps, eq_tests=False,
equationT=None, P=None, H2O_Liq=None, Fe3Fet_Liq=None, iterations=30):

    '''
    Evaluates all possible Ol-liq pairs for temperature, and calculates
    equilibrium tests

    Parameters
    ----------------

    ol_comps: pandas.DataFrame
        Panda DataFrame of Ol compositions with column headings SiO2_Ol, CaO_Ol etc.

    liq_comps: pandas.DataFrame
        Panda DataFrame of liq compositions with column headings SiO2_Liq etc.

    equationT: str
        T_Beatt93_ol (P-dependent, H2O_independent)
        T_Beatt93_ol_HerzCorr (P-dependent, H2O_independent)
        T_Put2008_eq21 (P-dependent, H2O-dependent)
        T_Put2008_eq22 (P-dependent, H2O-dependent)
        T_Sisson1992 (P-dependent, H2O_independent)
        T_Pu2017 (P-independent, H2O_independent)
        T_Pu2021 (P-dependent, H2O_independent)

    P: float, int, pandas.Series, str  ("Solve")
        Pressure in kbar
        Only needed for P-sensitive thermometers.
        If enter P="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    H2O_Liq: optional.
        If None, uses H2O_Liq column from input.
        If int, float, pandas.Series, uses this instead of H2O_Liq Column


    iterations: int
        number of times to iterate temperature and H2O. Default 30.

    Returns
    -------
    pandas.DataFrame
        Temp (K) for all posible ol-liq matches, along with equilibrium
        tests, components and input mineral compositions. At the moment, doesnt average per olivine.. But could upon request.
    '''

    ol_comps_c=ol_comps.copy()
    liq_comps_c=liq_comps.copy()

    #Check valid equation for T
    try:
        func = Liquid_olivine_funcs_by_name[equationT]
    except KeyError:
        raise ValueError(f'{equationT} is not a valid equation for Ol-Liquid Hygrometry') from None
    sig=inspect.signature(func)

    # This is a duplication step, to pair up all possible mathes. First we give a unique identifier

    ol_comps_c['ID_Ol'] = ol_comps_c.index
    liq_comps_c['ID_liq'] = liq_comps_c.index.astype('float64')

    if "Sample_ID_Ol" not in ol_comps.columns:
        ol_comps_c['Sample_ID_Ol'] = ol_comps.index.astype('float64')

    if 'Sample_ID_liq' not in liq_comps.columns:
        liq_comps_c['Sample_ID_liq'] = liq_comps.index.astype('float64')

    # Uses mole fractions, not cation fractions
    if equationT == "T_Pu2017" or equationT == "T_Pu2021":
        # For liquid
        anhyd_mol_frac_Ni = calculate_anhydrous_mol_fractions_liquid_Ni(
                liq_comps=liq_comps_c)
        # For olivine
        ol_mol_frac_Ni = calculate_mol_fractions_olivine_ni(
                    ol_comps=ol_comps_c)
        # So names stay the same
        ol_cat_frac=pd.concat([ol_mol_frac_Ni, ol_comps_c], axis=1)
        anhyd_liq_frac =pd.concat([anhyd_mol_frac_Ni, liq_comps_c], axis=1)



    else:

        anhyd_liq_frac = calculate_anhydrous_cat_fractions_liquid(liq_comps=liq_comps_c)
        ol_cat_frac1 = calculate_cat_fractions_olivine(ol_comps=ol_comps_c)
        ol_cat_frac= pd.concat([ol_comps_c, ol_cat_frac1], axis=1)


    # This duplicates Ols, repeats liq1-liq1*N, liq2-liq2*N etc.
    DupOls = pd.DataFrame(np.repeat(ol_cat_frac.values, np.shape(
        anhyd_liq_frac)[0], axis=0))  # .astype('float64')
    DupOls.columns = ol_cat_frac.columns

    # This duplicates liquids like liq1-liq2-liq3 for liq1, liq1-liq2-liq3 for
    # liq2 etc.
    Dupliqs = pd.concat([anhyd_liq_frac ] * np.shape(ol_cat_frac)[0]).reset_index(drop=True)
    # Combines these merged liquids and liq dataframes
    Combo_Ol_liqs = pd.concat([Dupliqs, DupOls], axis=1)

    Combo_Ol_liqs_1 = Combo_Ol_liqs.copy()
    # Combo_Ols_liqs_1['K_Barth'] = Combo_Ols_liqs_1['Ab_liq'] / \
    #     Combo_Ols_liqs_1['Ab_Ol']
    LenCombo = str(np.shape(Combo_Ol_liqs)[0])

    LenOl=len(ol_comps)
    LenLiqs=len(liq_comps)
    print("Considering N=" + str(LenOl) + " Ol & N=" + str(LenLiqs) +" Liqs, which is a total of N="+ str(LenCombo) +
          " Liq-Ol pairs, be patient if this is >>1 million!")

    CalcT=calculate_ol_liq_temp(equationT=equationT, meltmatch=Combo_Ol_liqs, Fe3Fet_Liq=Fe3Fet_Liq,
    H2O_Liq=H2O_Liq, eq_tests=eq_tests, P=P)


    Fo=calculate_ol_fo(ol_comps=DupOls)

    CalcT.insert(1, 'Ol_Fo_Meas', Fo, )

    df_out=pd.concat([CalcT, Combo_Ol_liqs_1], axis=1)

    return df_out





def calculate_ol_liq_temp(*, equationT, liq_comps=None, ol_comps=None, meltmatch=None, P=None,
                          NiO_Ol_Mol=None, H2O_Liq=None, Fe3Fet_Liq=None, eq_tests=False):
    '''
    Olivine-liquid thermometers. Returns the temperature in Kelvin,
    along with calculations of Kd-Fe-Mg equilibrium tests.

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    ol_comps: pandas.DataFrame
        Olivine compositions with column headings SiO2_Ol, MgO_Ol etc.

    equationT: str
        T_Beatt93_ol (P-dependent, H2O_independent)
        T_Beatt93_ol_HerzCorr (P-dependent, H2O_independent)
        T_Put2008_eq21 (P-dependent, H2O-dependent)
        T_Put2008_eq22 (P-dependent, H2O-dependent)
        T_Sisson1992 (P-dependent, H2O_independent)
        T_Pu2017 (P-independent, H2O_independent) Eq 1
        T_Pu2021 (P-dependent, H2O_independent) - only applies P corr for 10-30 GPa

    P: float, int, pandas.Series, str  ("Solve")
        Pressure in kbar
        Only needed for P-sensitive thermometers.
        If enter P="Solve", returns a partial function
        Else, enter an integer, float, or panda series

    H2O_Liq: optional.
        If None, uses H2O_Liq column from input.
        If int, float, pandas.Series, uses this instead of H2O_Liq Column


    Returns
    -------
    pandas.core.series.Series
        Temperatures in kelvin.


    '''
# These are checks that our inputs are okay
    if NiO_Ol_Mol is not None:
        raise TypeError('Sorry, this functionality was lost in the latest diadfit version to allow for melt matching using Pu. Please enter oliivne and liquid compositions instead')


    try:
        func = Liquid_olivine_funcs_by_name[equationT]
    except KeyError:
        raise ValueError(f'{equationT} is not a valid equation') from None
    sig=inspect.signature(func)

    if meltmatch is not None:
            Liq_Ols=meltmatch


    else:
        if len(liq_comps)!=len(ol_comps):
            raise ValueError('Ol comps need to be same length as Liq comps. use a _matching function calculate_ol_liq_temp_matching instead if you want to consider all pairs')


        ol_comps_c=ol_comps.copy()
        liq_comps_c=liq_comps.copy()

        if isinstance(P, pd.Series):
            if len(P) != len(liq_comps):
                raise ValueError('The panda series entered for pressure isnt the same length '
                'as the dataframe of liquid compositions')
            if len(liq_comps) != len(ol_comps):
                raise ValueError('The panda series entered for olivine isnt the same length as for liquids')
        liq_comps_c = liq_comps.copy()
        ol_comps_c=ol_comps.copy()


        if H2O_Liq is not None:
            liq_comps_c['H2O_Liq'] = H2O_Liq
        if Fe3Fet_Liq is not None:
            liq_comps_c['Fe3Fet_Liq'] = Fe3Fet_Liq

# Allows different calculation scheme for Ni-bearing equations



        else:
        # Keiths spreadsheets dont use Cr2O3 and P2O5. So have set this to zero.
            liq_comps_c['Cr2O3_Liq']=0
            liq_comps_c['P2O5_Liq']=0
            ol_comps_c['Cr2O3_Ol']=0
            ol_comps_c['P2O5_Ol']=0
    # Now calculate cation fractions


        if equationT == "T_Pu2017" or equationT == "T_Pu2021" or equationT == "T_Pu2017_Mg":
            anhyd_mol_frac_Ni = calculate_anhydrous_mol_fractions_liquid_Ni(
                liq_comps=liq_comps_c)
            # Calculate the liquid mole fraction of NiO
            ol_mol_frac_Ni = calculate_mol_fractions_olivine_ni(
                    ol_comps=ol_comps_c)

            Liq_Ols = pd.concat([anhyd_mol_frac_Ni, ol_mol_frac_Ni, ol_comps_c, liq_comps_c], axis=1)


        else:


            anhyd_cat_frac = calculate_anhydrous_cat_fractions_liquid(liq_comps=liq_comps_c)
            ol_cat_frac = calculate_cat_fractions_olivine(ol_comps=ol_comps_c)
            Liq_Ols = pd.concat([anhyd_cat_frac, ol_cat_frac, ol_comps_c], axis=1)




    # This performs extra calculation steps for Beattie equations
    if equationT == "T_Put2008_eq22" or equationT == "T_Put2008_eq21" or \
    equationT == "T_Beatt93_ol" or equationT == "T_Beatt93_ol_HerzCorr" or equationT=="T_Put2008_eq19":



        Liq_Ols['DMg_Meas'] = Liq_Ols['Mg_Ol_cat_frac'].astype(float) /Liq_Ols['Mg_Liq_cat_frac'].astype(float)
        Liq_Ols['CNML'] = (Liq_Ols['Mg_Liq_cat_frac'] + Liq_Ols['Fet_Liq_cat_frac'] +
                                Liq_Ols['Ca_Liq_cat_frac'] + Liq_Ols['Mn_Liq_cat_frac'])
        Liq_Ols['CSiO2L'] = Liq_Ols['Si_Liq_cat_frac']
        Liq_Ols['NF'] = (7 / 2) * np.log(1 - Liq_Ols['Al_Liq_cat_frac']
                                            ) + 7 * np.log(1 - Liq_Ols['Ti_Liq_cat_frac'])
        Liq_Ols['Den_Beat93'] = 52.05 / 8.3144 + 2 * np.log(Liq_Ols['DMg_Meas']) + 2 * np.log(
            1.5 * Liq_Ols['CNML']) + 2 * np.log(3 * Liq_Ols['CSiO2L']) - Liq_Ols['NF']

    if equationT == "T_Sisson1992":
        Liq_Ols['KdMg_TSG1992'] = (Liq_Ols['Mg_Ol_cat_frac'] /
            (Liq_Ols['Mg_Liq_cat_frac'] *
                (Liq_Ols['Si_Liq_cat_frac']**(0.5))))


# Checks if P-dependent function you have entered a P
    if sig.parameters['P'].default is not None:
        if P is None:
            raise ValueError(f'{equationT} requires you to enter P')
    else:
        if P is not None:
            print('Youve selected a P-independent function')


    kwargs = {name: Liq_Ols[name] for name, p in sig.parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}

    if isinstance(P, str) or P is None:
        if P == "Solve":
            T_K = partial(func, **kwargs)
        if P is None:
            T_K=func(**kwargs)

    else:

        T_K=func(P, **kwargs)

    if eq_tests is False:

        KdFeMg_Meas = (
            ((Liq_Ols['FeOt_Ol'] / 71.844) / (Liq_Ols['MgO_Ol'] / 40.3044)) /
            ((Liq_Ols['FeOt_Liq'] * (1 - Liq_Ols['Fe3Fet_Liq']
                                            ) / 71.844) / (Liq_Ols['MgO_Liq'] / 40.3044))
        )
        df = pd.DataFrame(
            data={'T_K_calc': T_K, 'Kd (Fe-Mg) Meas': KdFeMg_Meas})
        return df

    if eq_tests is True:
        if NiO_Ol_Mol is not None:
            raise Exception(
                'No olivine composition, so cannot calculate equilibrium test. Set eq_tests=False')
        if P is None:
            P = 1
            print(
                'You have not selected a pressure, so we have calculated Toplis Kd at 1kbar')
        ol_fo = (Liq_Ols['MgO_Ol'] / 40.3044) / \
            ((Liq_Ols['MgO_Ol'] / 40.3044) + Liq_Ols['FeOt_Ol'] / 71.844)
        KdFeMg_Meas = (
            ((Liq_Ols['FeOt_Ol'] / 71.844) / (Liq_Ols['MgO_Ol'] / 40.3044)) /
            ((Liq_Ols['FeOt_Liq'] * (1 - Liq_Ols['Fe3Fet_Liq']
                                         ) / 71.844) / (Liq_Ols['MgO_Liq'] / 40.3044))
        )
        Kd_func = partial(calculate_toplis2005_kd, SiO2_mol=Liq_Ols['SiO2_Liq_mol_frac'], Na2O_mol=Liq_Ols[
                          'Na2O_Liq_mol_frac'], K2O_mol=Liq_Ols['Na2O_Liq_mol_frac'], P=P, H2O=Liq_Ols['H2O_Liq'], T=T_K)
        Kd_Toplis_calc = Kd_func(ol_fo)

        DeltaKd_Toplis = KdFeMg_Meas - Kd_Toplis_calc
        DeltaKd_Roeder = KdFeMg_Meas - 0.3
        DeltaKd_Matzen = KdFeMg_Meas - 0.34
        DeltaKd_Shea= KdFeMg_Meas-0.335
        df = pd.DataFrame(data={'T_K_calc': T_K, 'Kd Meas': KdFeMg_Meas, 'Kd calc (Toplis)': Kd_Toplis_calc,
                                'ΔKd, Toplis (M-P)': DeltaKd_Toplis, 'ΔKd, Roeder (M-P)': DeltaKd_Roeder, 'ΔKd, Matzen (M-P)': DeltaKd_Matzen, 'ΔKd, Shea (M-P)': DeltaKd_Shea})
        df_out = pd.concat([df, Liq_Ols], axis=1)

        return df_out

## Functions for olivine-spinel thermometry equations

def T_Coogan2014(P=None, *, Cr_No_sp, Al2O3_Ol, Al2O3_Sp):
    '''
    Aluminum-in-olivine thermometer from Coogan et al. 2014. doi: 10.1016/j.chemgeo.2014.01.004
    Uses the Al2O3 content in Olivine, Al2O3 content of Spinel, and the Cr number of the spinel
    '''

    return (10000 / ((0.575) + 0.884 * Cr_No_sp -
            0.897 * np.log(Al2O3_Ol / Al2O3_Sp)))


def T_Wan2008(P=None, *, Cr_No_sp, Al2O3_Ol, Al2O3_Sp):
    '''
    Aluminum-in-olivine thermometer from Wan et al. (2008)  - doi: 10.2138/am.2008.2758
    Uses the Al2O3 content in Olivine, Al2O3 content of Spinel, and the Cr number of the spinel
    '''

    return (10000 / ((0.512) + 0.873 * Cr_No_sp -
            0.91 * np.log(Al2O3_Ol / Al2O3_Sp)))

##  Olivine-spinel thermometry function

def calculate_ol_sp_temp(ol_comps, sp_comps, equationT):
    ''' calculates temperatures from olivine-spinel pairs.


   Parameters
    -------
    ol_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Ol, MgO_Ol etc

    sp_comps: pandas.DataFrame
        spinel compositions with column headings SiO2_Sp, MgO_Sp etc

    equationT: str
        Equation choices:
            |   T_Wan2008
            |   T_Coogan2014

    Returns
    -------
    pandas series
       Temperature in K

    '''

    if len(sp_comps)!=len(ol_comps):
        raise ValueError('Ol comps need to be same length as Sp comps.we dont have a matching function yet if want to consider all pairs, but could make one on request!')



    combo = pd.concat([ol_comps, sp_comps], axis=1)
    Cr_No = (sp_comps['Cr2O3_Sp'] / 151.99) / \
        (sp_comps['Cr2O3_Sp'] / 151.9 + sp_comps['Al2O3_Sp'] / 101.96)


    combo.insert(1, "Cr_No_sp", Cr_No)
    if equationT == "T_Coogan2014":
        T_func = T_Coogan2014
    if equationT == "T_Wan2008":
        T_func = T_Wan2008
    kwargs = {name: combo[name] for name, p in inspect.signature(
        T_func).parameters.items() if p.kind == inspect.Parameter.KEYWORD_ONLY}
    T_K = T_func(**kwargs)
    df_T_K = pd.DataFrame(data={'T_K_calc' + str(equationT.strip('T_')): T_K})

    return T_K



import numpy as np
import matplotlib.pyplot as plt
from functools import partial
import inspect
import warnings as w
import numbers
import pandas as pd

from Thermobar.core import *










import os
def return_cali_dataset(model=None):
    """
    This function returns the calibration dataset for different models, often with mineral components returned.
    This allows you to make your own plots rather than using the generic_cali_plot() option.


    Parameters
    -------


    model: str
        AMPHIBOLE:
        Ridolfi2021:  Ridolfi et al. (2021)
        Putirka2016:  Putirka (2016)
        Mutch2016: Mutch et al. (2016)
        Zhang2017: Zhang et al. (2017)

        CPX:
        Putirka2008: Putirka (2008) - entire database for Cpx-Liq,
        (not used for all equations).
        Masotta2013: Masotta et al. (2013)
        Neave2017:  Neave and Putirka (2017) for Cpx-Liq
        Brugman2019: Brugman and Till 2019
        Petrelli2020:  Petrelli et al. (2020) for Cpx and Cpx-Liq
        Wang2021: Wang et al. (2021) Cpx-only,but contains Liq compositions too.
        Jorgenson2022: Jorgenson et al. (2022) Cpx-Liq and Cpx-only,

        PLAGIOCLASE
        Waters2015: Waters and Lange (2015) for plag-liq hygrometry
        Masotta2019: Masotta et al. (2019) plag-liq hygrometry

        LIQUID
        Shea2022: Shea et al. (2022) for Ol MgO thermometry and Kd values
        HelzThornber87: Helz and Thornber (1987).

    """
    import pkg_resources

    # Check model is supported
    sup_models = ['Ridolfi2021', 'Zhang2017', 'Putirka2016', 'Mutch2016', 'Shea2022', 'Wang2021',
              'Petrelli2020', 'Putirka2008', 'Neave2017', 'Brugman2019', 'Masotta2013',
              'Jorgenson2022', 'Waters2015', 'Masotta2019', 'HelzThornber87']


    if model not in sup_models:
        raise ValueError(f"The model '{model}' is not supported yet.")






    model_to_filename = {
        "Zhang2017": "Zhang17_Cali_input.csv",
        "Ridolfi2021": "Ridolfi_Cali_input.csv",
    "Putirka2016": "Putirka16_Cali_input.csv",
    "Mutch2016": 'Mutch_Cali_input.csv',
    "Shea2022":"Shea2022_Cali_input.csv",
    "Wang2021": "Wang21_Cali_input.csv",

    "Petrelli2020": 'Petrelli20_Cali_input.csv',
    "Putirka2008": 'Putirka2008_Cali_input.csv',

    "Neave2017": 'NeavePutirka_2017_Cali_input.csv',
    "Brugman2019": 'Brugman_2019_Cali_input.csv',
    "Masotta2013": 'Masotta_2013_Cali_input.csv',
    "Jorgenson2022":'Jorgenson2022_Cali_input.csv',
    "Waters2015":'Waters_Lange2015_Cali_input.csv',
    "Masotta2019": 'Masotta2019_Cali_input.csv',
    'HelzThornber87': 'HelzThornber_1987_Cali_Liq.csv'}

    # Get the file name for the given model

    file_name = model_to_filename.get(model)



    if file_name:
        # Construct the relative path to the CSV file
        csv_file_relative_path = file_name

        # Get the absolute path to the CSV file within the package
        csv_file_path = pkg_resources.resource_filename('Thermobar', csv_file_relative_path)

        # Debugging: Print the resolved file path
        print(f"Resolved CSV file path: {csv_file_path}")

        # Check if the file exists at the resolved path
        if not os.path.exists(csv_file_path):
            print(f"File not found at path: {csv_file_path}")

        # Now you can use csv_file_path with pandas or other libraries
        Cali_input = pd.read_csv(csv_file_path)
    else:
        raise ValueError(f"Model '{model}' not recognized")

    # check model is supported.
    sup_models = ['Ridolfi2021', 'Zhang2017', 'Putirka2016', 'Mutch2016', 'Shea2022', 'Wang2021', 'Petrelli2020', 'Putirka2008', 'Neave2017',
                  'Brugman2019', 'Masotta2013', 'Jorgenson2022', 'Waters2015', 'Masotta2019']

    # Check if the provided model is in the list of supported models
    if model not in sup_models:
        raise ValueError(f"Invalid model: {model}. Supported models with inbuilt calibration datasets are {', '.join(sup_models)}")




    if model=="Zhang2017":

        Cali_input['F_Amp']=0
        Cali_input['Cl_Amp']=0

    if model=="Putirka2016":

        Cali_input['F_Amp']=0
        Cali_input['Cl_Amp']=0




    if model=="Masotta2019":
        cat_plags = calculate_cat_fractions_plagioclase(plag_comps=Cali_input)
        Cali_input['XAn']=cat_plags['An_Plag']
        Cali_input['XAb']=cat_plags['Ab_Plag']
    return Cali_input




def generic_cali_plot(df, model=None, x=None, y=None, P_kbar=None, T_K=None, figsize=(7, 5),
 shape_cali='o', mfc_cali='white', mec_cali='k', ms_cali=5,
 shape_data='^', mfc_data='red', alpha_cali=1, alpha_data=1, mec_data='k',
 ms_data=10, order="cali bottom", save_fig=False, fig_name=None,  dpi=200):

    """
    This function plots your compositions amongst the calibration dataset for a variety of models where we could
    obtain the exact calibration dataset. see model for option.

    Parameters
    -------

    df: pandas DataFrame
        dataframe of your compositions, e.g. a dataframe of Cpx composition

    x and y: str
        What you want to plotted against each other. E.g. x="SiO2_Cpx", y="Al2O3_Cpx"

    model: str
        AMPHIBOLE:
        Ridolfi2021:  Ridolfi et al. (2012)
        Putirka2016:  Putirka (2016)
        Mutch2016: Mutch et al. (2016)
        Zhang2017: Zhang et al. (2017)


        CPX:
        Putirka2008: Putirka (2008) - entire database for Cpx-Liq,
        (not used for all equations).
        Masotta2013: Masotta et al. (2013)
        Neave2017:  Neave and Putirka (2017) for Cpx-Liq
        Brugman2019: Brugman and Till 2019
        Petrelli2020:  Petrelli et al. (2020) for Cpx and Cpx-Liq
        Wang2021: Wang et al. (2021) Cpx-only,but contains Liq compositions too.
        Jorgenson2022: Jorgenson et al. (2022) Cpx-Liq and Cpx-only,

        PLAGIOCLASE
        Waters2015: Waters and Lange (2015) for plag-liq hygrometry
        Masotta2019: Masotta et al. (2019) plag-liq hygrometry

        LIQUID
        Shea2022: Shea et al. (2022) for Ol MgO thermometry and Kd values



    P_kbar and T_K: pd.Series
        if you want to plot calculated pressure and temperature against the calibration range,
        specify those panda series here, and it will append them onto your dataframe

    Plot customization options.
    -------


    figsize: (x, y)
        figure size

    order: 'cali top' or 'cali bottom':
        whether cali or user data is plotted ontop. default, cali data bottom

    shape_cali, shape_data:
        matplotlib symbol, e.g. shape_cali='o' and shape_data='^' are the defaults

    ms_cali, ms_data:
        marker size for cali and user data

    mfc_cali, mfc_data:
        marker face color for cali and data (can be different)

    mec_cali, mfc_data:
        marker edge color for cali and data.

    alpha_cali, alpha_data:
        transparency of symbols

    save_fig: bool
        if True, saves figure
        Uses fig_name to give filename,  dpi to give dpi


    """

    #df_c=df.copy()
    if model=="Ridolfi2021" or model=="Putirka2016" or model=="Mutch2016" or model=="Zhang2017":
        Amp_sites=calculate_sites_ridolfi(amp_comps=df)
        df_c=pd.concat([df, Amp_sites], axis=1)


    elif model=="Petrelli2020" or model=="Wang2021" or model=="Putirka2008" or model=="Neave2017" or model=="Brugman2019" or model=="Masotta2013" or model=="Jorgenson2022":
        if "Jd" not in df:
            Cpx_sites=calculate_clinopyroxene_components(cpx_comps=df)
            df_c=Cpx_sites
        else:
            df_c=df
    else:

        df_c=df


    if P_kbar is not None:
        df_c['P_kbar']=P_kbar
    if T_K is not None:
        df_c['T_K']=T_K

    # Now get the cali dataset

    Cali_input=return_cali_dataset(model=model)




    if x not in df_c:
        print(df_c.columns)
        raise TypeError('x variable no present in input dataframe. Choose one of the columns printed above instead')
    if y not in df_c:
        print(df_c.columns)
        raise TypeError('y variable no present in input dataframe')
    if x not in Cali_input:
        print(Cali_input.columns)
        raise TypeError('x variable no present in calibration dataframe')
    if y not in Cali_input:
        print(Cali_input.columns)
        raise TypeError('y variable no present in calibration dataframe')




    fig, (ax1) = plt.subplots(1, 1, figsize=figsize)

    if order=="cali top":
        zorder_cali=0
        zorder_data=5
    if order=="cali bottom":
        zorder_cali=5
        zorder_data=0
    ax1.plot(df_c[x], df_c[y], shape_data,
    mfc=mfc_data, mec=mec_data, ms=ms_data, alpha=alpha_data, label='User Data', zorder=zorder_cali)

    ax1.plot(Cali_input[x], Cali_input[y], shape_cali,
    mfc=mfc_cali, mec=mec_cali, ms=ms_cali, alpha=alpha_cali, label=model, zorder=zorder_data)

    xlabel=x.replace('_', ' ')
    ylabel=y.replace('_', ' ')
    ax1.legend()
    ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel)

    if save_fig is True:
        fig.savefig(fig_name, dpi=200)

    return fig


